# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = TestCase base class definitions
# scapy.contrib.status = library


import abc
from collections import defaultdict

from scapy.utils import make_lined_table, SingleConversationSocket
from scapy.supersocket import SuperSocket
from scapy.contrib.automotive.scanner.graph import _Edge
from scapy.contrib.automotive.ecu import EcuState, EcuResponse
from scapy.error import Scapy_Exception


# Typing imports
from typing import (
    Any,
    Union,
    List,
    Optional,
    Dict,
    Tuple,
    Set,
    Callable,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from scapy.contrib.automotive.scanner.configuration import AutomotiveTestCaseExecutorConfiguration  # noqa: E501


# type definitions
_SocketUnion = Union[SuperSocket, SingleConversationSocket]
_TransitionCallable = Callable[[_SocketUnion, "AutomotiveTestCaseExecutorConfiguration", Dict[str, Any]], bool]  # noqa: E501
_CleanupCallable = Callable[[_SocketUnion, "AutomotiveTestCaseExecutorConfiguration"], bool]  # noqa: E501
_TransitionTuple = Tuple[_TransitionCallable, Dict[str, Any], Optional[_CleanupCallable]]  # noqa: E501


class AutomotiveTestCaseABC(metaclass=abc.ABCMeta):
    """
    Base class for "TestCase" objects. In automotive scanners, these TestCase
    objects are used for individual tasks, for example enumerating over one
    kind of functionality of the protocol. It is also possible, that
    these TestCase objects execute complex tests on an ECU.
    The TestCaseExecuter object has a list of TestCases. The executer
    manipulates a device under test (DUT), to enter a certain state. In this
    state, the TestCase object gets executed.
    """

    _supported_kwargs = {}  # type: Dict[str, Tuple[Any, Optional[Callable[[Any], bool]]]]  # noqa: E501
    _supported_kwargs_doc = ""

    @abc.abstractmethod
    def has_completed(self, state):
        # type: (EcuState) -> bool
        """
        Tells if this TestCase was executed for a certain state
        :param state: State of interest
        :return: True, if TestCase was executed in the questioned state
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def pre_execute(self,
                    socket,  # type: _SocketUnion
                    state,  # type: EcuState
                    global_configuration  # type: AutomotiveTestCaseExecutorConfiguration  # noqa: E501
                    ):  # type: (...) -> None
        """
        Will be executed previously to ``execute``. This function can be used
        to manipulate the configuration passed to execute.

        :param socket: Socket object with the connection to a DUT
        :param state: Current state of the DUT
        :param global_configuration: Configuration of the TestCaseExecutor
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        """
        Executes this TestCase for a given state

        :param socket: Socket object with the connection to a DUT
        :param state: Current state of the DUT
        :param kwargs: Local configuration of the TestCasesExecutor
        :return:
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def post_execute(self,
                     socket,  # type: _SocketUnion
                     state,  # type: EcuState
                     global_configuration  # type: AutomotiveTestCaseExecutorConfiguration  # noqa: E501
                     ):  # type: (...) -> None
        """
        Will be executed subsequently to ``execute``. This function can be used
        for additional evaluations after the ``execute``.

        :param socket: Socket object with the connection to a DUT
        :param state: Current state of the DUT
        :param global_configuration: Configuration of the TestCaseExecutor
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def show(self, dump=False, filtered=True, verbose=False):
        # type: (bool, bool, bool) -> Optional[str]
        """
        Shows results of TestCase

        :param dump: If True, the results will be returned; If False, the
                     results will be printed.
        :param filtered: If True, the negative responses will be filtered
                         dynamically.
        :param verbose: If True, additional information will be provided.
        :return: test results of TestCase if parameter ``dump`` is True,
                 else ``None``
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def completed(self):
        # type: () -> bool
        """
        Tells if this TestCase is completely executed
        :return: True, if TestCase is completely executed
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def supported_responses(self):
        # type: () -> List[EcuResponse]
        """
        Tells the supported responses in TestCase
        :return: The list of supported responses
        """
        raise NotImplementedError


class AutomotiveTestCase(AutomotiveTestCaseABC):
    """ Base class for TestCases"""

    _description = "AutomotiveTestCase"
    _supported_kwargs = AutomotiveTestCaseABC._supported_kwargs
    _supported_kwargs_doc = AutomotiveTestCaseABC._supported_kwargs_doc

    def __init__(self):
        # type: () -> None
        self._state_completed = defaultdict(bool)  # type: Dict[EcuState, bool]

    def has_completed(self, state):
        # type: (EcuState) -> bool
        return self._state_completed[state]

    @classmethod
    def check_kwargs(cls, kwargs):
        # type: (Dict[str, Any]) -> None
        for k, v in kwargs.items():
            if k not in cls._supported_kwargs.keys():
                raise Scapy_Exception(
                    "Keyword-Argument %s not supported for %s" %
                    (k, cls.__name__))
            ti, vf = cls._supported_kwargs[k]
            if ti is not None and not isinstance(v, ti):
                raise Scapy_Exception(
                    "Keyword-Value '%s' is not instance of type %s" %
                    (k, str(ti)))
            if vf is not None and not vf(v):
                raise Scapy_Exception(
                    "Validation Error: '%s: %s' is not in the allowed "
                    "value range" % (k, str(v))
                )

    @property
    def completed(self):
        # type: () -> bool
        return all(v for _, v in self._state_completed.items())

    @property
    def scanned_states(self):
        # type: () -> Set[EcuState]
        """
        Helper function to get all scanned states
        :return: all scanned states
        """
        return set(self._state_completed.keys())

    def pre_execute(self, socket, state, global_configuration):
        # type: (_SocketUnion, EcuState, AutomotiveTestCaseExecutorConfiguration) -> None  # noqa: E501
        pass

    def execute(self, socket, state, **kwargs):
        # type: (_SocketUnion, EcuState, Any) -> None
        raise NotImplementedError()

    def post_execute(self, socket, state, global_configuration):
        # type: (_SocketUnion, EcuState, AutomotiveTestCaseExecutorConfiguration) -> None  # noqa: E501
        pass

    def _show_header(self, **kwargs):
        # type: (Any) -> str
        s = "\n\n" + "=" * (len(self._description) + 10) + "\n"
        s += " " * 5 + self._description + "\n"
        s += "-" * (len(self._description) + 10) + "\n"

        return s + "\n"

    def _show_state_information(self, **kwargs):
        # type: (Any) -> str
        completed = [(state, self._state_completed[state])
                     for state in self.scanned_states]
        return make_lined_table(
            completed, lambda x, y: ("Scan state completed", x, y),
            dump=True) or ""

    def show(self, dump=False, filtered=True, verbose=False):
        # type: (bool, bool, bool) -> Optional[str]

        s = self._show_header()

        if verbose:
            s += self._show_state_information()

        if dump:
            return s + "\n"
        else:
            print(s)
            return None


class TestCaseGenerator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_generated_test_case(self):
        # type: () -> Optional[AutomotiveTestCaseABC]
        raise NotImplementedError()


class StateGenerator(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_new_edge(self, socket, config):
        # type: (_SocketUnion, AutomotiveTestCaseExecutorConfiguration) -> Optional[_Edge]  # noqa: E501
        raise NotImplementedError

    @abc.abstractmethod
    def get_transition_function(self, socket, edge):
        # type: (_SocketUnion, _Edge) -> Optional[_TransitionTuple]
        """

        :param socket: Socket to target
        :param edge: Tuple of EcuState objects for the requested
                     transition function
        :return: Returns an optional tuple consisting of a transition function,
                 a keyword arguments dictionary for the transition function
                 and a cleanup function. Both functions
                 take a Socket and the TestCaseExecutor configuration as
                 arguments and return True if the execution was successful.
                 The first function is the state enter function, the second
                 function is a cleanup function
        """
        raise NotImplementedError



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
External link to matplotlib
"""

from scapy.error import log_loading

# Notice: this file must not be called before main.py, if started
# in interactive mode, because it needs to be called after the
# logger has been setup, to be able to print the warning messages

__all__ = [
    "Line2D",
    "MATPLOTLIB",
    "MATPLOTLIB_DEFAULT_PLOT_KARGS",
    "MATPLOTLIB_INLINED",
    "plt",
]

# MATPLOTLIB

try:
    from matplotlib import get_backend as matplotlib_get_backend
    from matplotlib import pyplot as plt
    from matplotlib.lines import Line2D
    MATPLOTLIB = 1
    if "inline" in matplotlib_get_backend():
        MATPLOTLIB_INLINED = 1
    else:
        MATPLOTLIB_INLINED = 0
    MATPLOTLIB_DEFAULT_PLOT_KARGS = {"marker": "+"}
# RuntimeError to catch gtk "Cannot open display" error
except (ImportError, RuntimeError) as ex:
    plt = None
    Line2D = None
    MATPLOTLIB = 0
    MATPLOTLIB_INLINED = 0
    MATPLOTLIB_DEFAULT_PLOT_KARGS = dict()
    log_loading.info("Can't import matplotlib: %s. Won't be able to plot.", ex)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
NetBIOS over TCP/IP

[RFC 1001/1002]
"""

import struct
from scapy.arch import get_if_addr
from scapy.base_classes import Net
from scapy.ansmachine import AnsweringMachine
from scapy.compat import bytes_encode
from scapy.config import conf

from scapy.packet import Packet, bind_bottom_up, bind_layers, bind_top_down
from scapy.fields import (
    BitEnumField,
    BitField,
    ByteEnumField,
    ByteField,
    FieldLenField,
    FlagsField,
    IPField,
    IntField,
    NetBIOSNameField,
    PacketListField,
    ShortEnumField,
    ShortField,
    StrFixedLenField,
    XShortField,
    XStrFixedLenField
)
from scapy.interfaces import _GlobInterfaceType
from scapy.sendrecv import sr1
from scapy.layers.inet import IP, UDP, TCP
from scapy.layers.l2 import Ether, SourceMACField

# Typing imports
from typing import (
    List,
    Union,
)


class NetBIOS_DS(Packet):
    name = "NetBIOS datagram service"
    fields_desc = [
        ByteEnumField("type", 17, {17: "direct_group"}),
        ByteField("flags", 0),
        XShortField("id", 0),
        IPField("src", "127.0.0.1"),
        ShortField("sport", 138),
        ShortField("len", None),
        ShortField("ofs", 0),
        NetBIOSNameField("srcname", ""),
        NetBIOSNameField("dstname", ""),
    ]

    def post_build(self, p, pay):
        p += pay
        if self.len is None:
            tmp_len = len(p) - 14
            p = p[:10] + struct.pack("!H", tmp_len) + p[12:]
        return p

#        ShortField("length",0),
#        ShortField("Delimiter",0),
#        ByteField("command",0),
#        ByteField("data1",0),
#        ShortField("data2",0),
#        ShortField("XMIt",0),
#        ShortField("RSPCor",0),
#        StrFixedLenField("dest","",16),
#        StrFixedLenField("source","",16),
#
#        ]
#

# NetBIOS


_NETBIOS_SUFFIXES = {
    0x4141: "workstation",
    0x4141 + 0x03: "messenger service",
    0x4141 + 0x200: "file server service",
    0x4141 + 0x10b: "domain master browser",
    0x4141 + 0x10c: "domain controller",
    0x4141 + 0x10e: "browser election service"
}

_NETBIOS_QRTYPES = {
    0x20: "NB",
    0x21: "NBSTAT"
}

_NETBIOS_QRCLASS = {
    1: "INTERNET"
}

_NETBIOS_RNAMES = {
    0xC00C: "Label String Pointer to QUESTION_NAME"
}

_NETBIOS_OWNER_MODE_TYPES = {
    0: "B node",
    1: "P node",
    2: "M node",
    3: "H node"
}

_NETBIOS_GNAMES = {
    0: "Unique name",
    1: "Group name"
}


class NBNSHeader(Packet):
    name = "NBNS Header"
    fields_desc = [
        ShortField("NAME_TRN_ID", 0),
        BitField("RESPONSE", 0, 1),
        BitField("OPCODE", 0, 4),
        FlagsField("NM_FLAGS", 0, 7, ["B",
                                      "res1",
                                      "res0",
                                      "RA",
                                      "RD",
                                      "TC",
                                      "AA"]),
        BitField("RCODE", 0, 4),
        ShortField("QDCOUNT", 0),
        ShortField("ANCOUNT", 0),
        ShortField("NSCOUNT", 0),
        ShortField("ARCOUNT", 0),
    ]

    def hashret(self):
        return b"NBNS" + struct.pack("!B", self.OPCODE)


# Name Query Request
# RFC1002 sect 4.2.12


class NBNSQueryRequest(Packet):
    name = "NBNS query request"
    fields_desc = [NetBIOSNameField("QUESTION_NAME", "windows"),
                   ShortEnumField("SUFFIX", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL", 0),
                   ShortEnumField("QUESTION_TYPE", 0x20, _NETBIOS_QRTYPES),
                   ShortEnumField("QUESTION_CLASS", 1, _NETBIOS_QRCLASS)]

    def mysummary(self):
        return "NBNSQueryRequest who has '\\\\%s'" % (
            self.QUESTION_NAME.decode(errors="backslashreplace")
        )


bind_layers(NBNSHeader, NBNSQueryRequest,
            OPCODE=0x0, NM_FLAGS=0x11, QDCOUNT=1)


# Name Query Response
# RFC1002 sect 4.2.13


class NBNS_ADD_ENTRY(Packet):
    fields_desc = [
        BitEnumField("G", 0, 1, _NETBIOS_GNAMES),
        BitEnumField("OWNER_NODE_TYPE", 00, 2,
                     _NETBIOS_OWNER_MODE_TYPES),
        BitEnumField("UNUSED", 0, 13, {0: "Unused"}),
        IPField("NB_ADDRESS", "127.0.0.1")
    ]


class NBNSQueryResponse(Packet):
    name = "NBNS query response"
    fields_desc = [NetBIOSNameField("RR_NAME", "windows"),
                   ShortEnumField("SUFFIX", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL", 0),
                   ShortEnumField("QUESTION_TYPE", 0x20, _NETBIOS_QRTYPES),
                   ShortEnumField("QUESTION_CLASS", 1, _NETBIOS_QRCLASS),
                   IntField("TTL", 0x493e0),
                   FieldLenField("RDLENGTH", None, length_of="ADDR_ENTRY"),
                   PacketListField("ADDR_ENTRY",
                                   [NBNS_ADD_ENTRY()], NBNS_ADD_ENTRY,
                                   length_from=lambda pkt: pkt.RDLENGTH)
                   ]

    def mysummary(self):
        if not self.ADDR_ENTRY or \
           not isinstance(self.ADDR_ENTRY[0], NBNS_ADD_ENTRY):
            return "NBNSQueryResponse"
        return "NBNSQueryResponse '\\\\%s' is at %s" % (
            self.RR_NAME.decode(errors="backslashreplace"),
            self.ADDR_ENTRY[0].NB_ADDRESS
        )

    def answers(self, other):
        return (
            isinstance(other, NBNSQueryRequest) and
            other.QUESTION_NAME == self.RR_NAME
        )


bind_layers(NBNSHeader, NBNSQueryResponse,  # RD+AA
            OPCODE=0x0, NM_FLAGS=0x50, RESPONSE=1, ANCOUNT=1)
for _flg in [0x58, 0x70, 0x78]:
    bind_bottom_up(NBNSHeader, NBNSQueryResponse,
                   OPCODE=0x0, NM_FLAGS=_flg, RESPONSE=1, ANCOUNT=1)


# Node Status Request
# RFC1002 sect 4.2.17

class NBNSNodeStatusRequest(NBNSQueryRequest):
    name = "NBNS status request"
    QUESTION_NAME = b"*" + b"\x00" * 14
    QUESTION_TYPE = 0x21

    def mysummary(self):
        return "NBNSNodeStatusRequest who has '\\\\%s'" % (
            self.QUESTION_NAME.decode(errors="backslashreplace")
        )


bind_bottom_up(NBNSHeader, NBNSNodeStatusRequest, OPCODE=0x0, NM_FLAGS=0, QDCOUNT=1)
bind_layers(NBNSHeader, NBNSNodeStatusRequest, OPCODE=0x0, NM_FLAGS=1, QDCOUNT=1)


# Node Status Response
# RFC1002 sect 4.2.18

class NBNSNodeStatusResponseService(Packet):
    name = "NBNS Node Status Response Service"
    fields_desc = [StrFixedLenField("NETBIOS_NAME", "WINDOWS         ", 15),
                   ByteEnumField("SUFFIX", 0, {0: "workstation",
                                               0x03: "messenger service",
                                               0x20: "file server service",
                                               0x1b: "domain master browser",
                                               0x1c: "domain controller",
                                               0x1e: "browser election service"
                                               }),
                   ByteField("NAME_FLAGS", 0x4),
                   ByteEnumField("UNUSED", 0, {0: "unused"})]

    def default_payload_class(self, payload):
        return conf.padding_layer


class NBNSNodeStatusResponse(Packet):
    name = "NBNS Node Status Response"
    fields_desc = [NetBIOSNameField("RR_NAME", "windows"),
                   ShortEnumField("SUFFIX", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL", 0),
                   ShortEnumField("RR_TYPE", 0x21, _NETBIOS_QRTYPES),
                   ShortEnumField("RR_CLASS", 1, _NETBIOS_QRCLASS),
                   IntField("TTL", 0),
                   ShortField("RDLENGTH", 83),
                   FieldLenField("NUM_NAMES", None, fmt="B",
                                 count_of="NODE_NAME"),
                   PacketListField("NODE_NAME",
                                   [NBNSNodeStatusResponseService()],
                                   NBNSNodeStatusResponseService,
                                   count_from=lambda pkt: pkt.NUM_NAMES),
                   SourceMACField("MAC_ADDRESS"),
                   XStrFixedLenField("STATISTICS", b"", 46)]

    def answers(self, other):
        return (
            isinstance(other, NBNSNodeStatusRequest) and
            other.QUESTION_NAME == self.RR_NAME
        )


bind_layers(NBNSHeader, NBNSNodeStatusResponse,
            OPCODE=0x0, NM_FLAGS=0x40, RESPONSE=1, ANCOUNT=1)


# Name Registration Request
# RFC1002 sect 4.2.2

class NBNSRegistrationRequest(Packet):
    name = "NBNS registration request"
    fields_desc = [
        NetBIOSNameField("QUESTION_NAME", "Windows"),
        ShortEnumField("SUFFIX", 0x4141, _NETBIOS_SUFFIXES),
        ByteField("NULL", 0),
        ShortEnumField("QUESTION_TYPE", 0x20, _NETBIOS_QRTYPES),
        ShortEnumField("QUESTION_CLASS", 1, _NETBIOS_QRCLASS),
        ShortEnumField("RR_NAME", 0xC00C, _NETBIOS_RNAMES),
        ShortEnumField("RR_TYPE", 0x20, _NETBIOS_QRTYPES),
        ShortEnumField("RR_CLASS", 1, _NETBIOS_QRCLASS),
        IntField("TTL", 0),
        ShortField("RDLENGTH", 6),
        BitEnumField("G", 0, 1, _NETBIOS_GNAMES),
        BitEnumField("OWNER_NODE_TYPE", 00, 2,
                     _NETBIOS_OWNER_MODE_TYPES),
        BitEnumField("UNUSED", 0, 13, {0: "Unused"}),
        IPField("NB_ADDRESS", "127.0.0.1")
    ]

    def mysummary(self):
        return self.sprintf("Register %G% %QUESTION_NAME% at %NB_ADDRESS%")


bind_bottom_up(NBNSHeader, NBNSRegistrationRequest, OPCODE=0x5)
bind_layers(NBNSHeader, NBNSRegistrationRequest,
            OPCODE=0x5, NM_FLAGS=0x11, QDCOUNT=1, ARCOUNT=1)


# Wait for Acknowledgement Response
# RFC1002 sect 4.2.16

class NBNSWackResponse(Packet):
    name = "NBNS Wait for Acknowledgement Response"
    fields_desc = [NetBIOSNameField("RR_NAME", "windows"),
                   ShortEnumField("SUFFIX", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL", 0),
                   ShortEnumField("RR_TYPE", 0x20, _NETBIOS_QRTYPES),
                   ShortEnumField("RR_CLASS", 1, _NETBIOS_QRCLASS),
                   IntField("TTL", 2),
                   ShortField("RDLENGTH", 2),
                   BitField("RDATA", 10512, 16)]  # 10512=0010100100010000


bind_layers(NBNSHeader, NBNSWackResponse,
            OPCODE=0x7, NM_FLAGS=0x40, RESPONSE=1, ANCOUNT=1)

# NetBIOS DATAGRAM HEADER


class NBTDatagram(Packet):
    name = "NBT Datagram Packet"
    fields_desc = [ByteField("Type", 0x10),
                   ByteField("Flags", 0x02),
                   ShortField("ID", 0),
                   IPField("SourceIP", "127.0.0.1"),
                   ShortField("SourcePort", 138),
                   ShortField("Length", None),
                   ShortField("Offset", 0),
                   NetBIOSNameField("SourceName", "windows"),
                   ShortEnumField("SUFFIX1", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL1", 0),
                   NetBIOSNameField("DestinationName", "windows"),
                   ShortEnumField("SUFFIX2", 0x4141, _NETBIOS_SUFFIXES),
                   ByteField("NULL2", 0)]

    def post_build(self, pkt, pay):
        if self.Length is None:
            length = len(pay) + 68
            pkt = pkt[:10] + struct.pack("!H", length) + pkt[12:]
        return pkt + pay


# SESSION SERVICE PACKETS


class NBTSession(Packet):
    name = "NBT Session Packet"
    MAXLENGTH = 0x3ffff
    fields_desc = [ByteEnumField("TYPE", 0, {0x00: "Session Message",
                                             0x81: "Session Request",
                                             0x82: "Positive Session Response",
                                             0x83: "Negative Session Response",
                                             0x84: "Retarget Session Response",
                                             0x85: "Session Keepalive"}),
                   BitField("RESERVED", 0x00, 7),
                   BitField("LENGTH", None, 17)]

    def post_build(self, pkt, pay):
        if self.LENGTH is None:
            length = len(pay) & self.MAXLENGTH
            pkt = pkt[:1] + struct.pack("!I", length)[1:]
        return pkt + pay

    def extract_padding(self, s):
        return s[:self.LENGTH], s[self.LENGTH:]

    @classmethod
    def tcp_reassemble(cls, data, *args, **kwargs):
        if len(data) < 4:
            return None
        length = struct.unpack("!I", data[:4])[0] & cls.MAXLENGTH
        if len(data) >= length + 4:
            return cls(data)


bind_bottom_up(UDP, NBNSHeader, dport=137)
bind_bottom_up(UDP, NBNSHeader, sport=137)
bind_top_down(UDP, NBNSHeader, sport=137, dport=137)

bind_bottom_up(UDP, NBTDatagram, dport=138)
bind_bottom_up(UDP, NBTDatagram, sport=138)
bind_top_down(UDP, NBTDatagram, sport=138, dport=138)

bind_bottom_up(TCP, NBTSession, dport=445)
bind_bottom_up(TCP, NBTSession, sport=445)
bind_bottom_up(TCP, NBTSession, dport=139)
bind_bottom_up(TCP, NBTSession, sport=139)
bind_layers(TCP, NBTSession, dport=139, sport=139)


_nbns_cache = conf.netcache.new_cache("nbns_cache", 300)


@conf.commands.register
def nbns_resolve(
    qname: str,
    iface: Union[_GlobInterfaceType, List[_GlobInterfaceType]] = None,
    raw: bool = False,
    timeout: int = 3,
    **kwargs,
) -> List[str]:
    """
    Perform a simple NBNS (NetBios Name Services) resolution with caching

    :param qname: the name to query
    :param iface: the interfaces to use. (default: all)
    :param raw: return the whole netbios packet (default False)
    :param timeout: seconds until timeout (per server)
    :raise TimeoutError: if no DNS servers were reached in time.
    """
    kwargs.setdefault("verbose", 0)

    # Unify types (for caching)
    qname = NBNSQueryRequest.QUESTION_NAME.any2i(None, qname)

    # Check cache
    cache_ident = qname + b"raw" if raw else b""
    result = _nbns_cache.get(cache_ident)
    if result:
        return result

    if iface is None:
        ifaces = [
            x
            for name, x in conf.ifaces.items()
            if x.is_valid() and name != conf.loopback_name
        ]
    elif isinstance(iface, list):
        ifaces = iface
    else:
        ifaces = [iface]

    # Builds a request for each broadcast address of each interface
    requests = []
    for iface in ifaces:
        for bdcst in conf.route.get_if_bcast(iface):
            if bdcst == "255.255.255.255":
                continue
            requests.append(
                IP(dst=bdcst) /
                UDP() /
                NBNSHeader() /
                NBNSQueryRequest(QUESTION_NAME=qname)
            )

    if not requests:
        return None

    # Perform requests, get the first response
    try:
        old_checkIPAddr = conf.checkIPaddr
        conf.checkIPaddr = False

        res = sr1(
            requests,
            timeout=timeout,
            first=True,
            **kwargs,
        )
    finally:
        conf.checkIPaddr = old_checkIPAddr

    if res is not None:
        if raw:
            # Raw
            result = res
        else:
            # Get IP
            result = [x.NB_ADDRESS for x in res.ADDR_ENTRY]
        if result:
            # Cache it
            _nbns_cache[cache_ident] = result
        return result
    else:
        raise TimeoutError


class NBNS_am(AnsweringMachine):
    function_name = "nbnsd"
    filter = "udp port 137"
    sniff_options = {"store": 0}

    def parse_options(self, server_name=None, from_ip=None, ip=None):
        """
        NBNS answering machine

        :param server_name: the netbios server name to match
        :param from_ip: an IP (can have a netmask) to filter on
        :param ip: the IP to answer with
        """
        self.ServerName = bytes_encode(server_name or "")
        self.ip = ip
        if isinstance(from_ip, str):
            self.from_ip = Net(from_ip)
        else:
            self.from_ip = from_ip

    def is_request(self, req):
        if self.from_ip and IP in req and req[IP].src not in self.from_ip:
            return False
        return NBNSQueryRequest in req and (
            not self.ServerName or
            req[NBNSQueryRequest].QUESTION_NAME.strip() == self.ServerName
        )

    def make_reply(self, req):
        # type: (Packet) -> Packet
        resp = Ether(
            dst=req[Ether].src,
            src=None if req[Ether].dst == "ff:ff:ff:ff:ff:ff" else req[Ether].dst,
        ) / IP(dst=req[IP].src) / UDP(
            sport=req.dport,
            dport=req.sport,
        )
        address = self.ip or get_if_addr(self.optsniff.get("iface", conf.iface))
        resp /= NBNSHeader() / NBNSQueryResponse(
            RR_NAME=self.ServerName or req.QUESTION_NAME,
            SUFFIX=req.SUFFIX,
            ADDR_ENTRY=[NBNS_ADD_ENTRY(NB_ADDRESS=address)]
        )
        resp.NAME_TRN_ID = req.NAME_TRN_ID
        return resp



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = HomePlugGP Layer
# scapy.contrib.status = loads


from scapy.packet import Packet, bind_layers
from scapy.fields import ByteEnumField, ByteField, FieldLenField, \
    MACField, PacketListField, ShortField, \
    StrFixedLenField, XIntField, PacketField \

# This layer extends HomePlug AV one
from scapy.contrib.homeplugav import HomePlugAV, QualcommTypeList

# Copyright (C) HomePlugGP Layer for Scapy by FlUxIuS (Sebastien Dudek)
# As HomePlug GreenPHY is a subset of HomePlug AV, that is why we use
# HomePlugAV layer as a base here.

HomePlugGPTypes = {0x6008: "CM_SET_KEY_REQ",
                   0x6009: "CM_SET_KEY_CNF",
                   0x6064: "CM_SLAC_PARM_REQ",
                   0x6065: "CM_SLAC_PARM_CNF",
                   0x606e: "CM_ATTEN_CHAR_IN",
                   0x606a: "CM_START_ATTEN_CHAR_IND",
                   0x606f: "CM_ATTEN_CHAR_RSP",
                   0x6076: "CM_MNBC_SOUND_IND",
                   0x607c: "CM_SLAC_MATCH_REQ",
                   0x607d: "CM_SLAC_MATCH_CNF",
                   0x6086: "CM_ATTENUATION_CHARACTERISTICS_MME"}

QualcommTypeList.update(HomePlugGPTypes)

HPGP_codes = {0x0: "Success"}

KeyType_list = {0x01: "NMK (AES-128)"}

######################################################################
# SLAC operations
######################################################################


class CM_SLAC_PARM_REQ(Packet):
    name = "CM_SLAC_PARM_REQ"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   StrFixedLenField("RunID", b"\x00" * 8, 8)]


class CM_SLAC_PARM_CNF(Packet):
    name = "CM_SLAC_PARM_CNF"
    fields_desc = [MACField("MSoundTargetMAC", "00:00:00:00:00:00"),
                   ByteField("NumberMSounds", 0x0),
                   ByteField("TimeOut", 0x0),
                   ByteField("ResponseType", 0x0),
                   MACField("ForwardingSTA", "00:00:00:00:00:00"),
                   ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   StrFixedLenField("RunID", b"\x00" * 8, 8)]


class HPGP_GROUP(Packet):
    name = "HPGP_GROUP"
    fields_desc = [ByteField("group", 0x0)]

    def extract_padding(self, p):
        return "", p


class VS_ATTENUATION_CHARACTERISTICS_MME(Packet):
    name = "VS_ATTENUATION_CHARACTERISTICS_MME"
    fields_desc = [MACField("EVMACAddress", "00:00:00:00:00:00"),
                   FieldLenField("NumberOfGroups", None,
                                 count_of="Groups", fmt="B"),
                   ByteField("NumberOfCarrierPerGroupe", 0),
                   StrFixedLenField("Reserved", b"\x00" * 7, 7),
                   PacketListField("Groups", "", HPGP_GROUP,
                                   length_from=lambda pkt: pkt.NumberOfGroups)]


class CM_ATTENUATION_CHARACTERISTICS_MME(Packet):
    name = "CM_ATTENUATION_CHARACTERISTICS_MME"
    fields_desc = [MACField("EVMACAddress", "00:00:00:00:00:00"),
                   FieldLenField("NumberOfGroups", None, count_of="Groups",
                                 fmt="B"),
                   ByteField("NumberOfCarrierPerGroupe", 0),
                   PacketListField("Groups", "", HPGP_GROUP,
                                   length_from=lambda pkt: pkt.NumberOfGroups)]


class CM_ATTEN_CHAR_IND(Packet):
    name = "CM_ATTEN_CHAR_IND"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   MACField("SourceAdress", "00:00:00:00:00:00"),
                   StrFixedLenField("RunID", b"\x00" * 8, 8),
                   StrFixedLenField("SourceID", b"\x00" * 17, 17),
                   StrFixedLenField("ResponseID", b"\x00" * 17, 17),
                   ByteField("NumberOfSounds", 0x0),
                   FieldLenField("NumberOfGroups", None, count_of="Groups",
                                 fmt="B"),
                   PacketListField("Groups", "", HPGP_GROUP,
                                   length_from=lambda pkt: pkt.NumberOfGroups)]


class CM_ATTEN_CHAR_RSP(Packet):
    name = "CM_ATTEN_CHAR_RSP"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   MACField("SourceAdress", "00:00:00:00:00:00"),
                   StrFixedLenField("RunID", b"\x00" * 8, 8),
                   StrFixedLenField("SourceID", b"\x00" * 17, 17),
                   StrFixedLenField("ResponseID", b"\x00" * 17, 17),
                   ByteEnumField("Result", 0x0, HPGP_codes)]


class SLAC_varfield(Packet):
    name = "SLAC_varfield"
    fields_desc = [StrFixedLenField("EVID", b"\x00" * 17, 17),
                   MACField("EVMAC", "00:00:00:00:00:00"),
                   StrFixedLenField("EVSEID", b"\x00" * 17, 17),
                   MACField("EVSEMAC", "00:00:00:00:00:00"),
                   StrFixedLenField("RunID", b"\x00" * 8, 8),
                   StrFixedLenField("RSVD", b"\x00" * 8, 8)]


class CM_SLAC_MATCH_REQ(Packet):
    name = "CM_SLAC_MATCH_REQ"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   FieldLenField("MatchVariableFieldLen", None,
                                 length_of="VariableField", fmt="<H"),
                   PacketField("VariableField",
                               SLAC_varfield(),
                               SLAC_varfield)]


class SLAC_varfield_cnf(Packet):
    name = "SLAC_varfield"
    fields_desc = [StrFixedLenField("EVID", b"\x00" * 17, 17),
                   MACField("EVMAC", "00:00:00:00:00:00"),
                   StrFixedLenField("EVSEID", b"\x00" * 17, 17),
                   MACField("EVSEMAC", "00:00:00:00:00:00"),
                   StrFixedLenField("RunID", b"\x00" * 8, 8),
                   StrFixedLenField("RSVD", b"\x00" * 8, 8),
                   StrFixedLenField("NetworkID", b"\x00" * 7, 7),
                   ByteField("Reserved", 0x0),
                   StrFixedLenField("NMK", b"\x00" * 16, 16)]


class CM_SLAC_MATCH_CNF(Packet):
    name = "CM_SLAC_MATCH_CNF"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   FieldLenField("MatchVariableFieldLen", None,
                                 length_of="VariableField", fmt="<H"),
                   PacketField("VariableField",
                               SLAC_varfield_cnf(),
                               SLAC_varfield_cnf)]


class CM_START_ATTEN_CHAR_IND(Packet):
    name = "CM_START_ATTEN_CHAR_IND"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   ByteField("NumberOfSounds", 0x0),
                   ByteField("TimeOut", 0x0),
                   ByteField("ResponseType", 0x0),
                   MACField("ForwardingSTA", "00:00:00:00:00:00"),
                   StrFixedLenField("RunID", b"\x00" * 8, 8)]


class CM_MNBC_SOUND_IND(Packet):
    name = "CM_MNBC_SOUND_IND"
    fields_desc = [ByteField("ApplicationType", 0x0),
                   ByteField("SecurityType", 0x0),
                   StrFixedLenField("SenderID", b"\x00" * 17, 17),
                   ByteField("Countdown", 0x0),
                   StrFixedLenField("RunID", b"\x00" * 8, 8),
                   StrFixedLenField("RSVD", b"\x00" * 8, 8),
                   StrFixedLenField("RandomValue", b"\x00" * 16, 16)]


######################################################################
# Set keys for GP
######################################################################


class CM_SET_KEY_REQ(Packet):
    name = "CM_SET_KEY_REQ"
    fields_desc = [ByteEnumField("KeyType", 0x0, KeyType_list),
                   XIntField("MyNonce", 0),
                   XIntField("YourNonce", 0),
                   ByteField("PID", 0),
                   ShortField("ProtoRunNumber", 0),
                   ByteField("ProtoMessNumber", 0),
                   ByteField("CCoCapability", 0),
                   StrFixedLenField("NetworkID", b"\x00" * 7, 7),
                   ByteField("NewEncKeySelect", 0),
                   StrFixedLenField("NewKey", b"\x00" * 16, 16)]


class CM_SET_KEY_CNF(Packet):
    name = "CM_SET_KEY_CNF"
    fields_desc = [ByteEnumField("Result", 0x0, HPGP_codes),
                   XIntField("MyNonce", 0),
                   XIntField("YourNonce", 0),
                   ByteField("PID", 0),
                   ShortField("ProtoRunNumber", 0),
                   ByteField("ProtoMessNumber", 0),
                   ByteField("CCoCapability", 0)]


# END #


bind_layers(HomePlugAV, VS_ATTENUATION_CHARACTERISTICS_MME, HPtype=0xA14E)
bind_layers(HomePlugAV, CM_SLAC_PARM_REQ, HPtype=0x6064)
bind_layers(HomePlugAV, CM_SLAC_PARM_CNF, HPtype=0x6065)
bind_layers(HomePlugAV, CM_START_ATTEN_CHAR_IND, HPtype=0x606a)
bind_layers(HomePlugAV, CM_ATTEN_CHAR_IND, HPtype=0x606e)
bind_layers(HomePlugAV, CM_ATTEN_CHAR_RSP, HPtype=0x606f)
bind_layers(HomePlugAV, CM_MNBC_SOUND_IND, HPtype=0x6076)
bind_layers(HomePlugAV, CM_SLAC_MATCH_REQ, HPtype=0x607c)
bind_layers(HomePlugAV, CM_SLAC_MATCH_CNF, HPtype=0x607d)
bind_layers(HomePlugAV, CM_SET_KEY_REQ, HPtype=0x6008)
bind_layers(HomePlugAV, CM_SET_KEY_CNF, HPtype=0x6009)
bind_layers(HomePlugAV, CM_ATTENUATION_CHARACTERISTICS_MME, HPtype=0x6086)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2017 Maxence Tury
#               2019 Romain Perez

"""
TLS 1.3 key exchange logic.
"""

import struct

from scapy.config import conf, crypto_validator
from scapy.error import log_runtime
from scapy.fields import (
    FieldLenField,
    IntField,
    PacketField,
    PacketLenField,
    PacketListField,
    ShortEnumField,
    ShortField,
    StrFixedLenField,
    StrLenField,
    XStrLenField,
)
from scapy.packet import Packet
from scapy.layers.tls.extensions import TLS_Ext_Unknown, _tls_ext
from scapy.layers.tls.cert import PrivKeyECDSA, PrivKeyRSA, PrivKeyEdDSA
from scapy.layers.tls.crypto.groups import (
    _tls_named_curves,
    _tls_named_ffdh_groups,
    _tls_named_groups,
    _tls_named_groups_generate,
    _tls_named_groups_import,
    _tls_named_groups_pubbytes,
)

if conf.crypto_valid:
    from cryptography.hazmat.primitives.asymmetric import ec
if conf.crypto_valid_advanced:
    from cryptography.hazmat.primitives.asymmetric import ed25519
    from cryptography.hazmat.primitives.asymmetric import ed448


class KeyShareEntry(Packet):
    """
    When building from scratch, we create a DH private key, and when
    dissecting, we create a DH public key. Default group is secp256r1.
    """
    __slots__ = ["privkey", "pubkey"]
    name = "Key Share Entry"
    fields_desc = [ShortEnumField("group", None, _tls_named_groups),
                   FieldLenField("kxlen", None, length_of="key_exchange"),
                   XStrLenField("key_exchange", "",
                                length_from=lambda pkt: pkt.kxlen)]

    def __init__(self, *args, **kargs):
        self.privkey = None
        self.pubkey = None
        super(KeyShareEntry, self).__init__(*args, **kargs)

    def do_build(self):
        """
        We need this hack, else 'self' would be replaced by __iter__.next().
        """
        tmp = self.explicit
        self.explicit = True
        b = super(KeyShareEntry, self).do_build()
        self.explicit = tmp
        return b

    @crypto_validator
    def create_privkey(self):
        """
        This is called by post_build() for key creation.
        """
        self.privkey = _tls_named_groups_generate(self.group)
        self.key_exchange = _tls_named_groups_pubbytes(self.privkey)

    def post_build(self, pkt, pay):
        if self.group is None:
            self.group = 23     # secp256r1

        if not self.key_exchange:
            try:
                self.create_privkey()
            except ImportError:
                pass

        if self.kxlen is None:
            self.kxlen = len(self.key_exchange)

        group = struct.pack("!H", self.group)
        kxlen = struct.pack("!H", self.kxlen)
        return group + kxlen + self.key_exchange + pay

    @crypto_validator
    def register_pubkey(self):
        self.pubkey = _tls_named_groups_import(
            self.group,
            self.key_exchange
        )

    def post_dissection(self, r):
        try:
            self.register_pubkey()
        except ImportError:
            pass

    def extract_padding(self, s):
        return "", s


class TLS_Ext_KeyShare_CH(TLS_Ext_Unknown):
    name = "TLS Extension - Key Share (for ClientHello)"
    fields_desc = [ShortEnumField("type", 0x33, _tls_ext),
                   ShortField("len", None),
                   FieldLenField("client_shares_len", None,
                                 length_of="client_shares"),
                   PacketListField("client_shares", [], KeyShareEntry,
                                   length_from=lambda pkt: pkt.client_shares_len)]  # noqa: E501

    def post_build(self, pkt, pay):
        if not self.tls_session.frozen:
            privshares = self.tls_session.tls13_client_privshares
            for kse in self.client_shares:
                if kse.privkey:
                    if _tls_named_groups[kse.group] in privshares:
                        pkt_info = pkt.firstlayer().summary()
                        log_runtime.info("TLS: group %s used twice in the same ClientHello [%s]", kse.group, pkt_info)  # noqa: E501
                        break
                    privshares[_tls_named_groups[kse.group]] = kse.privkey
        return super(TLS_Ext_KeyShare_CH, self).post_build(pkt, pay)

    def post_dissection(self, r):
        if not self.tls_session.frozen:
            for kse in self.client_shares:
                if kse.pubkey:
                    pubshares = self.tls_session.tls13_client_pubshares
                    if _tls_named_groups[kse.group] in pubshares:
                        pkt_info = r.firstlayer().summary()
                        log_runtime.info("TLS: group %s used twice in the same ClientHello [%s]", kse.group, pkt_info)  # noqa: E501
                        break
                    pubshares[_tls_named_groups[kse.group]] = kse.pubkey
        return super(TLS_Ext_KeyShare_CH, self).post_dissection(r)


class TLS_Ext_KeyShare_HRR(TLS_Ext_Unknown):
    name = "TLS Extension - Key Share (for HelloRetryRequest)"
    fields_desc = [ShortEnumField("type", 0x33, _tls_ext),
                   ShortField("len", None),
                   ShortEnumField("selected_group", None, _tls_named_groups)]


class TLS_Ext_KeyShare_SH(TLS_Ext_Unknown):
    name = "TLS Extension - Key Share (for ServerHello)"
    fields_desc = [ShortEnumField("type", 0x33, _tls_ext),
                   ShortField("len", None),
                   PacketField("server_share", None, KeyShareEntry)]

    def post_build(self, pkt, pay):
        if not self.tls_session.frozen and self.server_share.privkey:
            # if there is a privkey, we assume the crypto library is ok
            privshare = self.tls_session.tls13_server_privshare
            if len(privshare) > 0:
                pkt_info = pkt.firstlayer().summary()
                log_runtime.info("TLS: overwriting previous server key share [%s]", pkt_info)  # noqa: E501
            group_name = _tls_named_groups[self.server_share.group]
            privshare[group_name] = self.server_share.privkey

            if group_name in self.tls_session.tls13_client_pubshares:
                privkey = self.server_share.privkey
                pubkey = self.tls_session.tls13_client_pubshares[group_name]
                if group_name in _tls_named_ffdh_groups.values():
                    pms = privkey.exchange(pubkey)
                elif group_name in _tls_named_curves.values():
                    if group_name in ["x25519", "x448"]:
                        pms = privkey.exchange(pubkey)
                    else:
                        pms = privkey.exchange(ec.ECDH(), pubkey)
                self.tls_session.tls13_dhe_secret = pms
                self.tls_session.kx_group = group_name
        return super(TLS_Ext_KeyShare_SH, self).post_build(pkt, pay)

    def post_dissection(self, r):
        if not self.tls_session.frozen and self.server_share.pubkey:
            # if there is a pubkey, we assume the crypto library is ok
            pubshare = self.tls_session.tls13_server_pubshare
            if pubshare:
                pkt_info = r.firstlayer().summary()
                log_runtime.info("TLS: overwriting previous server key share [%s]", pkt_info)  # noqa: E501
            group_name = _tls_named_groups[self.server_share.group]
            pubshare[group_name] = self.server_share.pubkey

            if group_name in self.tls_session.tls13_client_privshares:
                pubkey = self.server_share.pubkey
                privkey = self.tls_session.tls13_client_privshares[group_name]
                if group_name in _tls_named_ffdh_groups.values():
                    pms = privkey.exchange(pubkey)
                elif group_name in _tls_named_curves.values():
                    if group_name in ["x25519", "x448"]:
                        pms = privkey.exchange(pubkey)
                    else:
                        pms = privkey.exchange(ec.ECDH(), pubkey)
                self.tls_session.tls13_dhe_secret = pms
                self.tls_session.kx_group = group_name
            elif group_name in self.tls_session.tls13_server_privshare:
                pubkey = self.tls_session.tls13_client_pubshares[group_name]
                privkey = self.tls_session.tls13_server_privshare[group_name]
                if group_name in _tls_named_ffdh_groups.values():
                    pms = privkey.exchange(pubkey)
                elif group_name in _tls_named_curves.values():
                    if group_name in ["x25519", "x448"]:
                        pms = privkey.exchange(pubkey)
                    else:
                        pms = privkey.exchange(ec.ECDH(), pubkey)
                self.tls_session.tls13_dhe_secret = pms
                self.tls_session.kx_group = group_name
        return super(TLS_Ext_KeyShare_SH, self).post_dissection(r)


_tls_ext_keyshare_cls = {1: TLS_Ext_KeyShare_CH,
                         2: TLS_Ext_KeyShare_SH}

_tls_ext_keyshare_hrr_cls = {2: TLS_Ext_KeyShare_HRR}


class Ticket(Packet):
    name = "Recommended Ticket Construction (from RFC 5077)"
    fields_desc = [StrFixedLenField("key_name", None, 16),
                   StrFixedLenField("iv", None, 16),
                   FieldLenField("encstatelen", None, length_of="encstate"),
                   StrLenField("encstate", "",
                               length_from=lambda pkt: pkt.encstatelen),
                   StrFixedLenField("mac", None, 32)]


class TicketField(PacketLenField):
    def m2i(self, pkt, m):
        if len(m) < 64:
            # Minimum ticket size is 64 bytes
            return conf.raw_layer(m)
        return self.cls(m)


class PSKIdentity(Packet):
    name = "PSK Identity"
    fields_desc = [FieldLenField("identity_len", None,
                                 length_of="identity"),
                   TicketField("identity", "", Ticket,
                               length_from=lambda pkt: pkt.identity_len),
                   IntField("obfuscated_ticket_age", 0)]

    def default_payload_class(self, payload):
        return conf.padding_layer


class PSKBinderEntry(Packet):
    name = "PSK Binder Entry"
    fields_desc = [FieldLenField("binder_len", None, fmt="B",
                                 length_of="binder"),
                   StrLenField("binder", "",
                               length_from=lambda pkt: pkt.binder_len)]

    def default_payload_class(self, payload):
        return conf.padding_layer


class TLS_Ext_PreSharedKey_CH(TLS_Ext_Unknown):
    # XXX define post_build and post_dissection methods
    name = "TLS Extension - Pre Shared Key (for ClientHello)"
    fields_desc = [ShortEnumField("type", 0x29, _tls_ext),
                   ShortField("len", None),
                   FieldLenField("identities_len", None,
                                 length_of="identities"),
                   PacketListField("identities", [], PSKIdentity,
                                   length_from=lambda pkt: pkt.identities_len),
                   FieldLenField("binders_len", None,
                                 length_of="binders"),
                   PacketListField("binders", [], PSKBinderEntry,
                                   length_from=lambda pkt: pkt.binders_len)]


class TLS_Ext_PreSharedKey_SH(TLS_Ext_Unknown):
    name = "TLS Extension - Pre Shared Key (for ServerHello)"
    fields_desc = [ShortEnumField("type", 0x29, _tls_ext),
                   ShortField("len", None),
                   ShortField("selected_identity", None)]


_tls_ext_presharedkey_cls = {1: TLS_Ext_PreSharedKey_CH,
                             2: TLS_Ext_PreSharedKey_SH}


# Util to find usable signature algorithms

# TLS 1.3 SignatureScheme is a subset of _tls_hash_sig
_tls13_usable_certificate_verify_algs = [
    # ECDSA algorithms
    0x0403, 0x0503, 0x0603,
    # RSASSA-PSS algorithms with public key OID rsaEncryption
    0x0804, 0x0805, 0x0806,
    # EdDSA algorithms
    0x0807, 0x0808,
]

_tls13_usable_certificate_signature_algs = [
    # RSASSA-PKCS1-v1_5 algorithms
    0x0401, 0x0501, 0x0601,
    # ECDSA algorithms
    0x0403, 0x0503, 0x0603,
    # EdDSA algorithms
    0x0807, 0x0808,
    # RSASSA-PSS algorithms with public key OID RSASSA-PSS
    0x0809, 0x080a, 0x080b,
    # Legacy algorithms
    0x0201, 0x0203,
]


def get_usable_tls13_sigalgs(li, key, location="certificateverify"):
    """
    From a list of proposed signature algorithms, this function returns a list of
    usable signature algorithms.
    The order of the signature algorithms in the list returned by the
    function matches the one of the proposal.
    """
    from scapy.layers.tls.keyexchange import _tls_hash_sig
    res = []
    if isinstance(key, PrivKeyRSA):
        kx = "rsa"
    elif isinstance(key, PrivKeyECDSA):
        kx = "ecdsa"
    elif isinstance(key, PrivKeyEdDSA):
        if isinstance(key.pubkey.pubkey, ed25519.Ed25519PublicKey):
            kx = "ed25519"
        elif isinstance(key.pubkey.pubkey, ed448.Ed448PublicKey):
            kx = "ed448"
        else:
            kx = "unknown"
    else:
        return res
    if location == "certificateverify":
        algs = _tls13_usable_certificate_verify_algs
    elif location == "certificatesignature":
        algs = _tls13_usable_certificate_signature_algs
    else:
        return res
    for c in li:
        if c in algs:
            sigalg = _tls_hash_sig[c]
            if "+" in sigalg:
                _, sig = sigalg.split('+')
            else:
                sig = sigalg
            if kx in sig:
                res.append(c)
    return res



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2018 Hao Zheng <haozheng10@gmail.com>

# scapy.contrib.description = Generic Network Virtualization Encapsulation (GENEVE)
# scapy.contrib.status = loads

"""
Geneve: Generic Network Virtualization Encapsulation

https://datatracker.ietf.org/doc/html/rfc8926
"""

import struct

from scapy.fields import BitField, XByteField, XShortEnumField, X3BytesField, StrLenField, PacketListField
from scapy.packet import Packet, bind_layers
from scapy.layers.inet import IP, UDP
from scapy.layers.inet6 import IPv6
from scapy.layers.l2 import Ether, ETHER_TYPES

CLASS_IDS = {0x0100: "Linux",
             0x0101: "Open vSwitch",
             0x0102: "Open Virtual Networking (OVN)",
             0x0103: "In-band Network Telemetry (INT)",
             0x0104: "VMware",
             0x0105: "Amazon.com, Inc.",
             0x0106: "Cisco Systems, Inc.",
             0x0107: "Oracle Corporation",
             0x0110: "Amazon.com, Inc.",
             0x0118: "IBM",
             0x0128: "Ericsson",
             0xFEFF: "Unassigned",
             0xFFFF: "Experimental"}


class GeneveOptions(Packet):
    name = "Geneve Options"
    fields_desc = [XShortEnumField("classid", 0x0000, CLASS_IDS),
                   XByteField("type", 0x00),
                   BitField("reserved", 0, 3),
                   BitField("length", None, 5),
                   StrLenField('data', '', length_from=lambda x: x.length * 4)]

    def extract_padding(self, s):
        return "", s

    def post_build(self, p, pay):
        if self.length is None:
            tmp_len = len(self.data) // 4
            p = p[:3] + struct.pack("!B", (p[3] & 0x3) | (tmp_len & 0x1f)) + p[4:]
        return p + pay


class GENEVE(Packet):
    name = "GENEVE"
    fields_desc = [BitField("version", 0, 2),
                   BitField("optionlen", None, 6),
                   BitField("oam", 0, 1),
                   BitField("critical", 0, 1),
                   BitField("reserved", 0, 6),
                   XShortEnumField("proto", 0x0000, ETHER_TYPES),
                   X3BytesField("vni", 0),
                   XByteField("reserved2", 0x00),
                   PacketListField("options", [], GeneveOptions,
                                   length_from=lambda pkt: pkt.optionlen * 4)]

    def post_build(self, p, pay):
        if self.optionlen is None:
            tmp_len = (len(p) - 8) // 4
            p = struct.pack("!B", (p[0] & 0xc0) | (tmp_len & 0x3f)) + p[1:]
        return p + pay

    def answers(self, other):
        if isinstance(other, GENEVE):
            if ((self.proto == other.proto) and (self.vni == other.vni)):
                return self.payload.answers(other.payload)
        else:
            return self.payload.answers(other)
        return 0

    def mysummary(self):
        return self.sprintf("GENEVE (vni=%GENEVE.vni%,"
                            "optionlen=%GENEVE.optionlen%,"
                            "proto=%GENEVE.proto%)")


bind_layers(UDP, GENEVE, dport=6081)
bind_layers(GENEVE, Ether, proto=0x6558)
bind_layers(GENEVE, IP, proto=0x0800)
bind_layers(GENEVE, IPv6, proto=0x86dd)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2017 Maxence Tury

"""
Stateless HKDF for TLS 1.3.
"""

import struct

from scapy.config import conf, crypto_validator
from scapy.layers.tls.crypto.pkcs1 import _get_hash

if conf.crypto_valid:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.kdf.hkdf import HKDF, HKDFExpand
    from cryptography.hazmat.primitives.hashes import Hash
    from cryptography.hazmat.primitives.hmac import HMAC


class TLS13_HKDF(object):
    @crypto_validator
    def __init__(self, hash_name="sha256"):
        self.hash = _get_hash(hash_name)

    @crypto_validator
    def extract(self, salt, ikm):
        h = self.hash
        if ikm is None:
            ikm = b"\x00" * h.digest_size
        # cryptography 47.0.0 added this as a public API
        if getattr(HKDF, "extract", None) is not None:
            return HKDF.extract(h, salt, ikm)
        else:
            hkdf = HKDF(h, h.digest_size, salt, None, default_backend())
            return hkdf._extract(ikm)

    @crypto_validator
    def expand(self, prk, info, L):
        h = self.hash
        hkdf = HKDFExpand(h, L, info, default_backend())
        return hkdf.derive(prk)

    @crypto_validator
    def expand_label(self, secret, label, hash_value, length):
        hkdf_label = struct.pack("!H", length)
        hkdf_label += struct.pack("B", 6 + len(label))
        hkdf_label += b"tls13 "
        hkdf_label += label
        hkdf_label += struct.pack("B", len(hash_value))
        hkdf_label += hash_value
        return self.expand(secret, hkdf_label, length)

    @crypto_validator
    def derive_secret(self, secret, label, messages):
        h = Hash(self.hash, backend=default_backend())
        h.update(messages)
        hash_messages = h.finalize()
        hash_len = self.hash.digest_size
        return self.expand_label(secret, label, hash_messages, hash_len)

    @crypto_validator
    def compute_verify_data(self, basekey, handshake_context):
        hash_len = self.hash.digest_size
        finished_key = self.expand_label(basekey, b"finished", b"", hash_len)

        h = Hash(self.hash, backend=default_backend())
        h.update(handshake_context)
        hash_value = h.finalize()

        hm = HMAC(finished_key, self.hash, default_backend())
        hm.update(hash_value)
        return hm.finalize()



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Wireshark extcap API utils
https://www.wireshark.org/docs/wsdg_html_chunked/ChCaptureExtcap.html
"""

import collections
import functools
import pathlib
import re
import subprocess

from scapy.config import conf
from scapy.consts import WINDOWS
from scapy.data import MTU
from scapy.error import warning
from scapy.interfaces import (
    network_name,
    resolve_iface,
    InterfaceProvider,
    NetworkInterface,
)
from scapy.packet import Packet
from scapy.supersocket import SuperSocket
from scapy.utils import PcapReader, _create_fifo, _open_fifo

# Typing
from typing import (
    cast,
    Any,
    Dict,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)


def _extcap_call(prog: str,
                 args: List[str],
                 format: Dict[str, List[str]],
                 ) -> Dict[str, List[Tuple[str, ...]]]:
    """
    Function used to call a program using the extcap format,
    then parse the results
    """
    p = subprocess.Popen(
        [prog] + args,
        # On Windows, we must be in the Wireshark/ folder.
        cwd=pathlib.Path(prog).parent.parent,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        text=True
    )
    data, err = p.communicate()
    if p.returncode != 0:
        raise OSError("%s returned with error code %s: %s" % (prog, p.returncode, err))
    res = collections.defaultdict(list)
    for ifa in data.split("\n"):
        ifa = ifa.strip()
        for keyword, values in format.items():
            if not ifa.startswith(keyword):
                continue

            def _match(val: str, ifa: str) -> str:
                m = re.search(r"{%s=([^}]*)}" % val, ifa)
                if m:
                    return m.group(1)
                return ""
            res[keyword].append(
                tuple(
                    [_match(val, ifa) for val in values]
                )
            )
            break
    return cast(Dict[str, List[Tuple[str, ...]]], res)


class _ExtcapNetworkInterface(NetworkInterface):
    """
    Extcap NetworkInterface
    """

    def get_extcap_config(self) -> Dict[str, Tuple[str, ...]]:
        """
        Return a list of available configuration options on an extcap interface
        """
        return _extcap_call(
            self.provider.cmdprog,  # type: ignore
            ["--extcap-interface", self.network_name, "--extcap-config"],
            {
                "arg": ["number", "call", "display", "default", "required"],
                "value": ["arg", "value", "display", "default"],
            },
        )

    def get_extcap_cmd(self, **kwarg: Dict[str, str]) -> List[str]:
        """
        Return the extcap command line options
        """
        cmds = []
        for x in self.get_extcap_config()["arg"]:
            key = x[1].strip("-").replace("-", "_")
            if key in kwarg:
                # Apply argument
                cmds += [x[1], str(kwarg[key])]
            else:
                # Apply default
                if x[4] == "true":  # required
                    raise ValueError(
                        "Missing required argument: '%s' on iface %s." % (
                            key,
                            self.network_name,
                        )
                    )
                elif not x[3] or x[3] == "false":  # no default (or false)
                    continue
                if x[3] == "true":
                    cmds += [x[1]]
                else:
                    cmds += [x[1], x[3]]
        return cmds


class _ExtcapSocket(SuperSocket):
    """
    Read packets at layer 2 using an extcap command
    """

    nonblocking_socket = True

    @staticmethod
    def select(sockets: List[SuperSocket],
               remain: Optional[float] = None) -> List[SuperSocket]:
        return sockets

    def __init__(self, *_: Any, **kwarg: Any) -> None:
        cmdprog = kwarg.pop("cmdprog")
        iface = kwarg.pop("iface", None)
        if iface is None:
            raise NameError("Must select an interface for a extcap socket !")
        iface = resolve_iface(iface)
        if not isinstance(iface, _ExtcapNetworkInterface):
            raise ValueError("Interface should be an _ExtcapNetworkInterface")
        args = iface.get_extcap_cmd(**kwarg)
        iface = network_name(iface)
        self.outs = None  # extcap sockets can't write
        # open fifo
        fifo, fd = _create_fifo()
        args = ["--extcap-interface", iface, "--capture", "--fifo", fifo] + args
        self.proc = subprocess.Popen(
            [cmdprog] + args,
        )
        self.fd = _open_fifo(fd)
        self.reader = PcapReader(self.fd)  # type: ignore
        self.ins = self.reader  # type: ignore

    def recv(self, x: int = MTU, **kwargs: Any) -> Packet:
        return self.reader.recv(x, **kwargs)

    def close(self) -> None:
        self.proc.kill()
        self.proc.wait(timeout=2)
        SuperSocket.close(self)
        self.fd.close()


class _ExtcapInterfaceProvider(InterfaceProvider):
    """
    Interface provider made to hook on a extcap binary
    """

    headers = ("Index", "Name", "Address")
    header_sort = 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.cmdprog = kwargs.pop("cmdprog")
        super(_ExtcapInterfaceProvider, self).__init__(*args, **kwargs)

    def load(self) -> Dict[str, NetworkInterface]:
        data: Dict[str, NetworkInterface] = {}
        try:
            interfaces = _extcap_call(
                self.cmdprog,
                ["--extcap-interfaces"],
                {"interface": ["value", "display"]},
            )["interface"]
        except OSError as ex:
            warning(
                "extcap %s failed to load: %s",
                self.name,
                str(ex).strip().split("\n")[-1]
            )
            return {}
        for netw_name, name in interfaces:
            _index = re.search(r".*(\d+)", name)
            if _index:
                index = int(_index.group(1)) + 100
            else:
                index = 100
            if_data = {
                "name": name,
                "network_name": netw_name,
                "description": name,
                "index": index,
            }
            data[netw_name] = _ExtcapNetworkInterface(self, if_data)
        return data

    def _l2listen(self, _: Any) -> Type[SuperSocket]:
        return functools.partial(_ExtcapSocket, cmdprog=self.cmdprog)  # type: ignore

    def _l3socket(self, *_: Any) -> NoReturn:
        raise ValueError("Only sniffing is available for an extcap provider !")

    _l2socket = _l3socket  # type: ignore

    def _is_valid(self, dev: NetworkInterface) -> bool:
        return True

    def _format(self,
                dev: NetworkInterface,
                **kwargs: Any
                ) -> Tuple[Union[str, List[str]], ...]:
        """Returns a tuple of the elements used by show()"""
        return (str(dev.index), dev.name, dev.network_name)


def load_extcap() -> None:
    """
    Load extcap folder from wireshark and populate Scapy's providers.

    Additional interfaces should appear in conf.ifaces.
    """
    if WINDOWS:
        pattern = re.compile(r"^[^.]+(?:\.bat|\.exe)?$")
    else:
        pattern = re.compile(r"^[^.]+(?:\.sh)?$")
    for fld in conf.prog.extcap_folders:
        root = pathlib.Path(fld)
        for _cmdprog in root.glob("*"):
            if not _cmdprog.is_file() or not pattern.match(_cmdprog.name):
                continue
            cmdprog = str((root / _cmdprog).absolute())
            # success
            provname = pathlib.Path(cmdprog).name.rsplit(".", 1)[0]

            class _prov(_ExtcapInterfaceProvider):
                name = provname

            conf.ifaces.register_provider(
                functools.partial(_prov, cmdprog=cmdprog)  # type: ignore
            )



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = EtherIP
# scapy.contrib.status = loads

from scapy.fields import BitField
from scapy.packet import Packet, bind_layers
from scapy.layers.inet import IP
from scapy.layers.l2 import Ether


class EtherIP(Packet):
    name = "EtherIP / RFC 3378"
    fields_desc = [BitField("version", 3, 4),
                   BitField("reserved", 0, 12)]


bind_layers(IP, EtherIP, frag=0, proto=0x61)
bind_layers(EtherIP, Ether)



#! /usr/bin/env python

# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Damian Zaręba <damianzrb@zohomail.eu>

# scapy.contrib.description = AUTOSAR PDU packets handling package.
# scapy.contrib.status = loads
from typing import Tuple, Optional
from scapy.layers.inet import UDP
from scapy.fields import XIntField, PacketListField, LenField
from scapy.packet import Packet, bind_bottom_up


class PDU(Packet):
    """
    Single PDU Packet inside PDUTransport list.
    Contains ID and payload length, and later - raw load.
    It's free to interpret using bind_layers/bind_bottom_up method

    Based off this document:

    https://www.autosar.org/fileadmin/standards/classic/22-11/AUTOSAR_SWS_IPDUMultiplexer.pdf # noqa: E501
    """
    name = 'PDU'
    fields_desc = [
        XIntField('pdu_id', 0),
        LenField('pdu_payload_len', None, fmt="I")]

    def extract_padding(self, s):
        # type: (bytes) -> Tuple[bytes, Optional[bytes]]
        return s[:self.pdu_payload_len], s[self.pdu_payload_len:]


class PDUTransport(Packet):
    """
    Packet representing PDUTransport containing multiple PDUs
    """
    name = 'PDUTransport'
    fields_desc = [
        PacketListField("pdus", [PDU()], PDU)
    ]


bind_bottom_up(UDP, PDUTransport, dport=60000)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

from scapy.fields import FieldLenField, FieldListField, StrFixedLenField, \
    ByteField, ShortField, FlagsField, XByteField, PacketListField
from scapy.packet import Packet, bind_layers
from scapy.contrib.automotive.obd.packet import OBD_Packet
from scapy.contrib.automotive.obd.services import OBD_S09


# See https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_09
# for further information
# IID = Information IDentification

class OBD_S09_PR_Record(Packet):
    fields_desc = [
        XByteField("iid", 0),
    ]


class OBD_S09_PR(Packet):
    name = "Infotype IDs"
    fields_desc = [
        PacketListField("data_records", [], OBD_S09_PR_Record)
    ]

    def answers(self, other):
        return isinstance(other, OBD_S09) \
            and all(r.iid in other.iid for r in self.data_records)


class OBD_IID00(OBD_Packet):
    name = "IID_00_Service9SupportedInformationTypes"
    fields_desc = [
        FlagsField('supported_iids', 0, 32, [
            'IID20',
            'IID1F',
            'IID1E',
            'IID1D',
            'IID1C',
            'IID1B',
            'IID1A',
            'IID19',
            'IID18',
            'IID17',
            'IID16',
            'IID15',
            'IID14',
            'IID13',
            'IID12',
            'IID11',
            'IID10',
            'IID0F',
            'IID0E',
            'IID0D',
            'IID0C',
            'IID0B',
            'IID0A',
            'IID09',
            'IID08',
            'IID07',
            'IID06',
            'IID05',
            'IID04',
            'IID03',
            'IID02',
            'IID01'
        ])
    ]


class _OBD_IID_MessageCount(OBD_Packet):
    fields_desc = [
        ByteField('message_count', 0)
    ]


class OBD_IID01(_OBD_IID_MessageCount):
    name = "IID_01_VinMessageCount"


class OBD_IID03(_OBD_IID_MessageCount):
    name = "IID_03_CalibrationIdMessageCount"


class OBD_IID05(_OBD_IID_MessageCount):
    name = "IID_05_CalibrationVerificationNumbersMessageCount"


class OBD_IID07(_OBD_IID_MessageCount):
    name = "IID_07_InUsePerformanceTrackingMessageCount"


class OBD_IID09(_OBD_IID_MessageCount):
    name = "IID_09_EcuNameMessageCount"


class OBD_IID02(OBD_Packet):
    name = "IID_02_VehicleIdentificationNumber"
    fields_desc = [
        FieldLenField('count', None, count_of='vehicle_identification_numbers',
                      fmt='B'),
        FieldListField('vehicle_identification_numbers', [],
                       StrFixedLenField('', b'', 17),
                       count_from=lambda pkt: pkt.count)
    ]


class OBD_IID04(OBD_Packet):
    name = "IID_04_CalibrationId"
    fields_desc = [
        FieldLenField('count', None, count_of='calibration_identifications',
                      fmt='B'),
        FieldListField('calibration_identifications', [],
                       StrFixedLenField('', b'', 16),
                       count_from=lambda pkt: pkt.count)
    ]


class OBD_IID06(OBD_Packet):
    name = "IID_06_CalibrationVerificationNumbers"
    fields_desc = [
        FieldLenField('count', None,
                      count_of='calibration_verification_numbers', fmt='B'),
        FieldListField('calibration_verification_numbers', [],
                       StrFixedLenField('', b'', 4),
                       count_from=lambda pkt: pkt.count)
    ]


class OBD_IID08(OBD_Packet):
    name = "IID_08_InUsePerformanceTracking"
    fields_desc = [
        FieldLenField('count', None, count_of='data', fmt='B'),
        FieldListField('data', [],
                       ShortField('', 0),
                       count_from=lambda pkt: pkt.count)
    ]


class OBD_IID0A(OBD_Packet):
    name = "IID_0A_EcuName"
    fields_desc = [
        FieldLenField('count', None, count_of='ecu_names', fmt='B'),
        FieldListField('ecu_names', [],
                       StrFixedLenField('', b'', 20),
                       count_from=lambda pkt: pkt.count)
    ]


class OBD_IID0B(OBD_Packet):
    name = "IID_0B_InUsePerformanceTrackingForCompressionIgnitionVehicles"
    fields_desc = [
        FieldLenField('count', None, count_of='data', fmt='B'),
        FieldListField('data', [],
                       ShortField('', 0),
                       count_from=lambda pkt: pkt.count)
    ]


bind_layers(OBD_S09_PR_Record, OBD_IID00, iid=0x00)
bind_layers(OBD_S09_PR_Record, OBD_IID01, iid=0x01)
bind_layers(OBD_S09_PR_Record, OBD_IID02, iid=0x02)
bind_layers(OBD_S09_PR_Record, OBD_IID03, iid=0x03)
bind_layers(OBD_S09_PR_Record, OBD_IID04, iid=0x04)
bind_layers(OBD_S09_PR_Record, OBD_IID05, iid=0x05)
bind_layers(OBD_S09_PR_Record, OBD_IID06, iid=0x06)
bind_layers(OBD_S09_PR_Record, OBD_IID07, iid=0x07)
bind_layers(OBD_S09_PR_Record, OBD_IID08, iid=0x08)
bind_layers(OBD_S09_PR_Record, OBD_IID09, iid=0x09)
bind_layers(OBD_S09_PR_Record, OBD_IID0A, iid=0x0A)
bind_layers(OBD_S09_PR_Record, OBD_IID0B, iid=0x0B)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information

"""
Commonly used structures shared across Scapy
"""

import ctypes


class bpf_insn(ctypes.Structure):
    """"The BPF instruction data structure"""
    _fields_ = [("code", ctypes.c_ushort),
                ("jt", ctypes.c_ubyte),
                ("jf", ctypes.c_ubyte),
                ("k", ctypes.c_int)]


class bpf_program(ctypes.Structure):
    """"Structure for BIOCSETF"""
    _fields_ = [('bf_len', ctypes.c_int),
                ('bf_insns', ctypes.POINTER(bpf_insn))]


class sock_fprog(ctypes.Structure):
    """"Structure for SO_ATTACH_FILTER"""
    _fields_ = [('len', ctypes.c_ushort),
                ('filter', ctypes.POINTER(bpf_insn))]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Santiago Hernandez Ramos <shramos@protonmail.com>

# scapy.contrib.description = Message Queuing Telemetry Transport (MQTT)
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers
from scapy.fields import (
    BitEnumField,
    ByteEnumField,
    ByteField,
    ConditionalField,
    FieldLenField,
    FieldListField,
    PacketListField,
    ShortField,
    StrLenField,
)
from scapy.layers.inet import TCP
from scapy.error import Scapy_Exception
from scapy.compat import orb, chb
from scapy.volatile import RandNum
from scapy.config import conf


# CUSTOM FIELDS
# source: http://stackoverflow.com/a/43717630
class VariableFieldLenField(FieldLenField):
    def addfield(self, pkt, s, val):
        val = self.i2m(pkt, val)
        data = []
        while val:
            if val > 127:
                data.append(val & 127)
                val //= 128
            else:
                data.append(val)
                lastoffset = len(data) - 1
                data = b"".join(chb(val | (0 if i == lastoffset else 128))
                                for i, val in enumerate(data))
                return s + data
            if len(data) > 3:
                raise Scapy_Exception("%s: malformed length field" %
                                      self.__class__.__name__)
        # If val is None / 0
        return s + b"\x00"

    def getfield(self, pkt, s):
        value = 0
        for offset, curbyte in enumerate(s):
            curbyte = orb(curbyte)
            value += (curbyte & 127) * (128 ** offset)
            if curbyte & 128 == 0:
                return s[offset + 1:], value
            if offset > 2:
                raise Scapy_Exception("%s: malformed length field" %
                                      self.__class__.__name__)

    def randval(self):
        return RandVariableFieldLen()


class RandVariableFieldLen(RandNum):
    def __init__(self):
        RandNum.__init__(self, 0, 268435455)


# LAYERS
CONTROL_PACKET_TYPE = {
    1: 'CONNECT',
    2: 'CONNACK',
    3: 'PUBLISH',
    4: 'PUBACK',
    5: 'PUBREC',
    6: 'PUBREL',
    7: 'PUBCOMP',
    8: 'SUBSCRIBE',
    9: 'SUBACK',
    10: 'UNSUBSCRIBE',
    11: 'UNSUBACK',
    12: 'PINGREQ',
    13: 'PINGRESP',
    14: 'DISCONNECT',
    15: 'AUTH'  # Added in v5.0
}


QOS_LEVEL = {
    0: 'At most once delivery',
    1: 'At least once delivery',
    2: 'Exactly once delivery'
}


# source: http://stackoverflow.com/a/43722441
class MQTT(Packet):
    name = "MQTT fixed header"
    fields_desc = [
        BitEnumField("type", 1, 4, CONTROL_PACKET_TYPE),
        BitEnumField("DUP", 0, 1, {0: 'Disabled',
                                   1: 'Enabled'}),
        BitEnumField("QOS", 0, 2, QOS_LEVEL),
        BitEnumField("RETAIN", 0, 1, {0: 'Disabled',
                                      1: 'Enabled'}),
        # Since the size of the len field depends on the next layer, we need
        # to "cheat" with the length_of parameter and use adjust parameter to
        # calculate the value.
        VariableFieldLenField("len", None, length_of="len",
                              adjust=lambda pkt, x: len(pkt.payload),),
    ]


PROTOCOL_LEVEL = {
    3: 'v3.1',
    4: 'v3.1.1',
    5: 'v5.0'
}


class MQTTConnect(Packet):
    name = "MQTT connect"
    fields_desc = [
        FieldLenField("length", None, length_of="protoname"),
        StrLenField("protoname", "",
                    length_from=lambda pkt: pkt.length),
        ByteEnumField("protolevel", 5, PROTOCOL_LEVEL),
        BitEnumField("usernameflag", 0, 1, {0: 'Disabled',
                                            1: 'Enabled'}),
        BitEnumField("passwordflag", 0, 1, {0: 'Disabled',
                                            1: 'Enabled'}),
        BitEnumField("willretainflag", 0, 1, {0: 'Disabled',
                                              1: 'Enabled'}),
        BitEnumField("willQOSflag", 0, 2, QOS_LEVEL),
        BitEnumField("willflag", 0, 1, {0: 'Disabled',
                                        1: 'Enabled'}),
        BitEnumField("cleansess", 0, 1, {0: 'Disabled',
                                         1: 'Enabled'}),
        BitEnumField("reserved", 0, 1, {0: 'Disabled',
                                        1: 'Enabled'}),
        ShortField("klive", 0),
        FieldLenField("clientIdlen", None, length_of="clientId"),
        StrLenField("clientId", "",
                    length_from=lambda pkt: pkt.clientIdlen),
        # Payload with optional fields depending on the flags
        ConditionalField(FieldLenField("wtoplen", None, length_of="willtopic"),
                         lambda pkt: pkt.willflag == 1),
        ConditionalField(StrLenField("willtopic", "",
                                     length_from=lambda pkt: pkt.wtoplen),
                         lambda pkt: pkt.willflag == 1),
        ConditionalField(FieldLenField("wmsglen", None, length_of="willmsg"),
                         lambda pkt: pkt.willflag == 1),
        ConditionalField(StrLenField("willmsg", "",
                                     length_from=lambda pkt: pkt.wmsglen),
                         lambda pkt: pkt.willflag == 1),
        ConditionalField(FieldLenField("userlen", None, length_of="username"),
                         lambda pkt: pkt.usernameflag == 1),
        ConditionalField(StrLenField("username", "",
                                     length_from=lambda pkt: pkt.userlen),
                         lambda pkt: pkt.usernameflag == 1),
        ConditionalField(FieldLenField("passlen", None, length_of="password"),
                         lambda pkt: pkt.passwordflag == 1),
        ConditionalField(StrLenField("password", "",
                                     length_from=lambda pkt: pkt.passlen),
                         lambda pkt: pkt.passwordflag == 1),
    ]


class MQTTDisconnect(Packet):
    name = "MQTT disconnect"
    fields_desc = []


RETURN_CODE = {
    0: 'Connection Accepted',
    1: 'Unacceptable protocol version',
    2: 'Identifier rejected',
    3: 'Server unavailable',
    4: 'Bad username/password',
    5: 'Not authorized'
}


class MQTTConnack(Packet):
    name = "MQTT connack"
    fields_desc = [
        ByteField("sessPresentFlag", 0),
        ByteEnumField("retcode", 0, RETURN_CODE),
        # this package has not payload
    ]


class MQTTPublish(Packet):
    name = "MQTT publish"
    fields_desc = [
        FieldLenField("length", None, length_of="topic"),
        StrLenField("topic", "",
                    length_from=lambda pkt: pkt.length),
        ConditionalField(ShortField("msgid", None),
                         lambda pkt: (pkt.underlayer.QOS == 1 or
                                      pkt.underlayer.QOS == 2)),
        StrLenField("value", "",
                    length_from=lambda pkt: pkt.underlayer.len - pkt.length - 2
                    if pkt.underlayer.QOS == 0 else
                    pkt.underlayer.len - pkt.length - 4)
    ]


class MQTTPuback(Packet):
    name = "MQTT puback"
    fields_desc = [
        ShortField("msgid", None),
    ]


class MQTTPubrec(Packet):
    name = "MQTT pubrec"
    fields_desc = [
        ShortField("msgid", None),
    ]


class MQTTPubrel(Packet):
    name = "MQTT pubrel"
    fields_desc = [
        ShortField("msgid", None),
    ]


class MQTTPubcomp(Packet):
    name = "MQTT pubcomp"
    fields_desc = [
        ShortField("msgid", None),
    ]


class MQTTTopic(Packet):
    name = "MQTT topic"
    fields_desc = [
        FieldLenField("length", None, length_of="topic"),
        StrLenField("topic", "", length_from=lambda pkt:pkt.length)
    ]

    def guess_payload_class(self, payload):
        return conf.padding_layer


class MQTTTopicQOS(MQTTTopic):
    fields_desc = MQTTTopic.fields_desc + [ByteEnumField("QOS", 0, QOS_LEVEL)]


class MQTTSubscribe(Packet):
    name = "MQTT subscribe"
    fields_desc = [
        ShortField("msgid", None),
        PacketListField("topics", [], pkt_cls=MQTTTopicQOS)
    ]


ALLOWED_RETURN_CODE = {
    0x00: 'Granted QoS 0',
    0x01: 'Granted QoS 1',
    0x02: 'Granted QoS 2',
    0x80: 'Unspecified error',
    0x83: 'Implementation specific error',
    0x87: 'Not authorized',
    0x8F: 'Topic Filter invalid',
    0x91: 'Packet Identifier in use',
    0x97: 'Quota exceeded',
    0x9E: 'Shared Subscriptions not supported',
    0xA1: 'Subscription Identifiers not supported',
    0xA2: 'Wildcard Subscriptions not supported',
}


class MQTTSuback(Packet):
    name = "MQTT suback"
    fields_desc = [
        ShortField("msgid", None),
        FieldListField("retcodes", None, ByteEnumField("", None, ALLOWED_RETURN_CODE))
    ]


class MQTTUnsubscribe(Packet):
    name = "MQTT unsubscribe"
    fields_desc = [
        ShortField("msgid", None),
        PacketListField("topics", [], pkt_cls=MQTTTopic)
    ]


class MQTTUnsuback(Packet):
    name = "MQTT unsuback"
    fields_desc = [
        ShortField("msgid", None)
    ]


# LAYERS BINDINGS

bind_layers(TCP, MQTT, sport=1883)
bind_layers(TCP, MQTT, dport=1883)
bind_layers(MQTT, MQTTConnect, type=1)
bind_layers(MQTT, MQTTConnack, type=2)
bind_layers(MQTT, MQTTPublish, type=3)
bind_layers(MQTT, MQTTPuback, type=4)
bind_layers(MQTT, MQTTPubrec, type=5)
bind_layers(MQTT, MQTTPubrel, type=6)
bind_layers(MQTT, MQTTPubcomp, type=7)
bind_layers(MQTT, MQTTSubscribe, type=8)
bind_layers(MQTT, MQTTSuback, type=9)
bind_layers(MQTT, MQTTUnsubscribe, type=10)
bind_layers(MQTT, MQTTUnsuback, type=11)
bind_layers(MQTT, MQTTDisconnect, type=14)
bind_layers(MQTTConnect, MQTT)
bind_layers(MQTTConnack, MQTT)
bind_layers(MQTTPublish, MQTT)
bind_layers(MQTTPuback, MQTT)
bind_layers(MQTTPubrec, MQTT)
bind_layers(MQTTPubrel, MQTT)
bind_layers(MQTTPubcomp, MQTT)
bind_layers(MQTTSubscribe, MQTT)
bind_layers(MQTTSuback, MQTT)
bind_layers(MQTTUnsubscribe, MQTT)
bind_layers(MQTTUnsuback, MQTT)
bind_layers(MQTTDisconnect, MQTT)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#               2015, 2016, 2017 Maxence Tury

"""
Aggregate some TLS crypto objects.
"""

from scapy.layers.tls.crypto.suites import *  # noqa: F401



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Answering machines.
"""

########################
#  Answering machines  #
########################

import abc
import functools
import threading
import socket
import warnings

from scapy.arch import get_if_addr
from scapy.config import conf
from scapy.sendrecv import sendp, sniff, AsyncSniffer
from scapy.packet import Packet
from scapy.plist import PacketList

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

_T = TypeVar("_T", Packet, PacketList)


class ReferenceAM(type):
    def __new__(cls,
                name,  # type: str
                bases,  # type: Tuple[type, ...]
                dct  # type: Dict[str, Any]
                ):
        # type: (...) -> Type['AnsweringMachine[_T]']
        obj = cast('Type[AnsweringMachine[_T]]',
                   super(ReferenceAM, cls).__new__(cls, name, bases, dct))
        try:
            import inspect
            obj.__signature__ = inspect.signature(  # type: ignore
                obj.parse_options
            )
        except (ImportError, AttributeError):
            pass
        if obj.function_name:
            func = lambda obj=obj, *args, **kargs: obj(*args, **kargs)()  # type: ignore  # noqa: E501
            # Inject signature
            func.__name__ = func.__qualname__ = obj.function_name
            func.__doc__ = obj.__doc__ or obj.parse_options.__doc__
            try:
                func.__signature__ = obj.__signature__  # type: ignore
            except (AttributeError):
                pass
            globals()[obj.function_name] = func
        return obj


class AnsweringMachine(Generic[_T], metaclass=ReferenceAM):
    function_name = ""
    filter = None  # type: Optional[str]
    sniff_options = {"store": 0}  # type: Dict[str, Any]
    sniff_options_list = ["store", "iface", "count", "promisc", "filter",
                          "type", "prn", "stop_filter", "opened_socket"]
    send_options = {"verbose": 0}  # type: Dict[str, Any]
    send_options_list = ["iface", "inter", "loop", "verbose", "socket"]
    send_function = staticmethod(sendp)

    def __init__(self, **kargs):
        # type: (Any) -> None
        self.mode = 0
        self.verbose = kargs.get("verbose", conf.verb >= 0)
        if self.filter:
            kargs.setdefault("filter", self.filter)
        kargs.setdefault("prn", self.reply)
        self.optam1 = {}  # type: Dict[str, Any]
        self.optam2 = {}  # type: Dict[str, Any]
        self.optam0 = {}  # type: Dict[str, Any]
        doptsend, doptsniff = self.parse_all_options(1, kargs)
        self.defoptsend = self.send_options.copy()
        self.defoptsend.update(doptsend)
        self.defoptsniff = self.sniff_options.copy()
        self.defoptsniff.update(doptsniff)
        self.optsend = {}   # type: Dict[str, Any]
        self.optsniff = {}   # type: Dict[str, Any]

    def __getattr__(self, attr):
        # type: (str) -> Any
        for dct in [self.optam2, self.optam1]:
            if attr in dct:
                return dct[attr]
        raise AttributeError(attr)

    def __setattr__(self, attr, val):
        # type: (str, Any) -> None
        mode = self.__dict__.get("mode", 0)
        if mode == 0:
            self.__dict__[attr] = val
        else:
            [self.optam1, self.optam2][mode - 1][attr] = val

    def parse_options(self):
        # type: () -> None
        pass

    def parse_all_options(self, mode, kargs):
        # type: (int, Any) -> Tuple[Dict[str, Any], Dict[str, Any]]
        sniffopt = {}  # type: Dict[str, Any]
        sendopt = {}  # type: Dict[str, Any]
        for k in list(kargs):  # use list(): kargs is modified in the loop
            if k in self.sniff_options_list:
                sniffopt[k] = kargs[k]
            if k in self.send_options_list:
                sendopt[k] = kargs[k]
            if k in self.sniff_options_list + self.send_options_list:
                del kargs[k]
        if mode != 2 or kargs:
            if mode == 1:
                self.optam0 = kargs
            elif mode == 2 and kargs:
                k = self.optam0.copy()
                k.update(kargs)
                self.parse_options(**k)
                kargs = k
            omode = self.__dict__.get("mode", 0)
            self.__dict__["mode"] = mode
            self.parse_options(**kargs)
            self.__dict__["mode"] = omode
        return sendopt, sniffopt

    def is_request(self, req):
        # type: (Packet) -> int
        return 1

    @abc.abstractmethod
    def make_reply(self, req):
        # type: (Packet) -> _T
        pass

    def send_reply(self, reply, send_function=None):
        # type: (_T, Optional[Callable[..., None]]) -> None
        if send_function:
            send_function(reply)
        else:
            self.send_function(reply, **self.optsend)

    def print_reply(self, req, reply):
        # type: (Packet, _T) -> None
        if isinstance(reply, PacketList):
            print("%s ==> %s" % (req.summary(),
                                 [res.summary() for res in reply]))
        else:
            print("%s ==> %s" % (req.summary(), reply.summary()))

    def reply(self, pkt, send_function=None, address=None):
        # type: (Packet, Optional[Callable[..., None]], Optional[Any]) -> None
        if not self.is_request(pkt):
            return
        if address:
            # Only on AnsweringMachineTCP
            reply = self.make_reply(pkt, address=address)  # type: ignore
        else:
            reply = self.make_reply(pkt)
        if not reply:
            return
        if send_function:
            self.send_reply(reply, send_function=send_function)
        else:
            # Retro-compability. Remove this if eventually
            self.send_reply(reply)
        if self.verbose:
            self.print_reply(pkt, reply)

    def run(self, *args, **kargs):
        # type: (Any, Any) -> None
        warnings.warn(
            "run() method deprecated. The instance is now callable",
            DeprecationWarning
        )
        self(*args, **kargs)

    def bg(self, *args, **kwargs):
        # type: (Any, Any) -> AsyncSniffer
        kwargs.setdefault("bg", True)
        self(*args, **kwargs)
        return self.sniffer

    def __call__(self, *args, **kargs):
        # type: (Any, Any) -> None
        bg = kargs.pop("bg", False)
        optsend, optsniff = self.parse_all_options(2, kargs)
        self.optsend = self.defoptsend.copy()
        self.optsend.update(optsend)
        self.optsniff = self.defoptsniff.copy()
        self.optsniff.update(optsniff)

        if bg:
            self.sniff_bg()
        else:
            try:
                self.sniff()
            except KeyboardInterrupt:
                print("Interrupted by user")

    def sniff(self):
        # type: () -> None
        sniff(**self.optsniff)

    def sniff_bg(self):
        # type: () -> None
        self.sniffer = AsyncSniffer(**self.optsniff)
        self.sniffer.start()


class AnsweringMachineTCP(AnsweringMachine[Packet]):
    """
    An answering machine that use the classic socket.socket to
    answer multiple TCP clients
    """
    TYPE = socket.SOCK_STREAM

    def parse_options(self, port=80, cls=conf.raw_layer):
        # type: (int, Type[Packet]) -> None
        self.port = port
        self.cls = cls

    def close(self):
        # type: () -> None
        pass

    def sniff(self):
        # type: () -> None
        from scapy.supersocket import StreamSocket
        ssock = socket.socket(socket.AF_INET, self.TYPE)
        try:
            ssock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except OSError:
            pass
        ssock.bind(
            (get_if_addr(self.optsniff.get("iface", conf.iface)), self.port))
        ssock.listen()
        sniffers = []
        try:
            while True:
                clientsocket, address = ssock.accept()
                print("%s connected" % repr(address))
                sock = StreamSocket(clientsocket, self.cls)
                optsniff = self.optsniff.copy()
                optsniff["prn"] = functools.partial(self.reply,
                                                    send_function=sock.send,
                                                    address=address)
                del optsniff["iface"]
                sniffer = AsyncSniffer(opened_socket=sock, **optsniff)
                sniffer.start()
                sniffers.append((sniffer, sock))
        finally:
            for (sniffer, sock) in sniffers:
                try:
                    sniffer.stop()
                except Exception:
                    pass
                sock.close()
            self.close()
            ssock.close()

    def sniff_bg(self):
        # type: () -> None
        self.sniffer = threading.Thread(target=self.sniff)  # type: ignore
        self.sniffer.start()

    def make_reply(self, req, address=None):
        # type: (Packet, Optional[Any]) -> Packet
        return req


class AnsweringMachineUDP(AnsweringMachineTCP):
    """
    An answering machine that use the classic socket.socket to
    answer multiple UDP clients
    """
    TYPE = socket.SOCK_DGRAM



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Operation, administration and maintenance (OAM)
# scapy.contrib.status = loads

"""
    Operation, administration and maintenance (OAM)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :author:    Sergey Matsievskiy, matsievskiysv@gmail.com

    :description:

        This module provides Scapy layers for the OAM protocol.

        normative references:
          - ITU-T Rec. G.8013/Y.1731 (08/2019) - Operation, administration and
            maintenance (OAM) functions and mechanisms for Ethernet-based
            networks (https://www.itu.int/rec/T-REC-G.8013)
          - ITU-T Rec. G.8031/Y.1342 (01/2015) - Ethernet linear protection
            switching (https://www.itu.int/rec/T-REC-G.8031)
          - ITU-T Rec. G.8032/Y.1344 (02/2022) - Ethernet ring protection
            switching (https://www.itu.int/rec/T-REC-G.8032)
"""

from scapy.fields import (
    BitEnumField,
    BitField,
    ByteField,
    ConditionalField,
    EnumField,
    FCSField,
    FlagsField,
    IntField,
    LenField,
    LongField,
    MACField,
    MultipleTypeField,
    NBytesField,
    OUIField,
    PacketField,
    PadField,
    PacketListField,
    ShortField,
    FieldListField,
)
from scapy.layers.l2 import Dot1Q
from scapy.packet import Packet, bind_layers
from binascii import crc32
import struct


class MepIdField(ShortField):
    """
    Short field with insignificant three leading bytes
    """

    def __init__(self, name, default):
        super(MepIdField, self).__init__(
            name, default & 0x1FFF if default is not None else default
        )


class MegId(Packet):
    """
    MEG ID
    """

    name = "MEG ID"

    fields_desc = [
        ByteField("resv", 1),
        ByteField("format", 0),
        MultipleTypeField(
            [
                (
                    LenField("length", 13, fmt="B"),
                    lambda p: p.format == 32,
                ),
                (
                    LenField("length", 15, fmt="B"),
                    lambda p: p.format == 33,
                )
            ],
            LenField("length", 45, fmt="B"),
        ),
        PadField(
            MultipleTypeField(
                [
                    (
                        FieldListField("values", [0] * 13,
                                       ByteField("value", 0),
                                       count_from=lambda pkt: pkt.length),
                        lambda x: x.format == 32,
                    ),
                    (
                        FieldListField("values", [0] * 15,
                                       ByteField("value", 0),
                                       count_from=lambda pkt: pkt.length),
                        lambda x: x.format == 33,
                    )
                ],
                NBytesField("values", 0, sz=45),
            ),
            45),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM_TLV(Packet):
    """
    OAM TLV
    """

    name = "OAM TLV"
    fields_desc = [ByteField("type", 1), LenField("length", None)]

    def extract_padding(self, s):
        return s[:self.length], s[self.length:]


class OAM_DATA_TLV(Packet):
    """
    OAM Data TLV
    """

    name = "OAM Data TLV"
    fields_desc = [ByteField("type", 3), LenField("length", None)]

    def extract_padding(self, s):
        return s[:self.length], s[self.length:]


class OAM_TEST_TLV(Packet):
    """
    OAM test TLV data
    """

    name = "OAM test TLV"

    fields_desc = [
        ByteField("type", 32),
        MultipleTypeField(
            [
                (
                    LenField("length", None, adjust=lambda l: l + 5),
                    lambda p: p.pat_type == 1 or p.pat_type == 3,
                )
            ],
            LenField("length", None, adjust=lambda l: l + 1),
        ),
        EnumField(
            "pat_type",
            0,
            {
                0: "Null signal without CRC-32",
                1: "Null signal with CRC-32",
                2: "PRBS 2^-31 - 1 without CRC-32",
                3: "PRBS 2^-31 - 1 with CRC-32",
            },
            fmt="B",
        ),
        ConditionalField(
            FCSField("crc", None, fmt="!I"),
            lambda p: p.pat_type == 1 or p.pat_type == 3,
        ),
    ]

    def do_dissect(self, s):
        if ord(s[3:4]) == 1 or ord(s[3:4]) == 3:
            # move crc to the end of packet
            length = struct.unpack("!H", s[1:3])[0]
            crc_end = 3 + length
            crc_start = crc_end - 4
            s1 = s[:crc_start]
            s2 = s[crc_start:crc_end]
            s3 = s[crc_end:]
            s = s1 + s3 + s2
        s = super(OAM_TEST_TLV, self).do_dissect(s)
        return s

    def post_build(self, p, pay):
        if ord(p[3:4]) == 1 or ord(p[3:4]) == 3:
            p1 = p
            p2 = pay[:-4]
            p3 = struct.pack("!I", crc32(p1 + p2) % (1 << 32))
            return p1 + p2 + p3
        else:
            return p + pay

    def extract_padding(self, s):
        if self.pat_type == 1 or self.pat_type == 3:
            # we already consumed crc
            return s[:self.length - 5], s[self.length - 5:]
        else:
            return s[:self.length - 1], s[self.length - 1:]


class OAM_LTM_TLV(Packet):
    """
    OAM LTM TLV data
    """

    name = "OAM LTM Egress ID TLV"

    fields_desc = [
        ByteField("type", 7),
        LenField("length", 8),
        LongField("egress_id", 0),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM_LTR_TLV(Packet):
    """
    OAM LTR TLV data
    """

    name = "OAM LTR Egress ID TLV"

    fields_desc = [
        ByteField("type", 8),
        LenField("length", 16),
        # NOTE: wireshark interprets this field as short+MAC
        LongField("last_egress_id", 0),
        # NOTE: wireshark interprets this field as short+MAC
        LongField("next_egress_id", 0),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM_LTR_IG_TLV(Packet):
    """
    OAM LTR TLV data
    """

    name = "OAM LTR Ingress TLV"

    fields_desc = [
        ByteField("type", 5),
        LenField("length", 7),
        ByteField("ingress_act", 0),
        MACField("ingress_mac", None),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM_LTR_EG_TLV(Packet):
    """
    OAM LTR TLV data
    """

    name = "OAM LTR Egress TLV"

    fields_desc = [
        ByteField("type", 6),
        LenField("length", 7),
        ByteField("egress_act", 0),
        MACField("egress_mac", None),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM_TEST_ID_TLV(Packet):
    """
    OAM Test ID TLV data
    """

    name = "OAM Test ID TLV"

    fields_desc = [
        ByteField("type", 36),
        LenField("length", 32),
        IntField("test_id", 0),
    ]

    def extract_padding(self, s):
        return b"", s


def guess_tlv_type(pkt, lst, cur, remain):
    if remain[0:1] == b'\x00':
        return None
    elif remain[0:1] == b'\x03':
        return OAM_DATA_TLV
    elif remain[0:1] == b'\x05':
        return OAM_LTR_IG_TLV
    elif remain[0:1] == b'\x06':
        return OAM_LTR_EG_TLV
    elif remain[0:1] == b'\x07':
        return OAM_LTM_TLV
    elif remain[0:1] == b'\x08':
        return OAM_LTR_TLV
    elif remain[0:1] == b'\x20':
        return OAM_TEST_TLV
    elif remain[0:1] == b'\x24':
        return OAM_TEST_ID_TLV
    else:
        return OAM_TLV


class PTP_TIMESTAMP(Packet):
    """
    PTP timestamp
    """

    # TODO: should be a part of PTP module
    name = "PTP timestamp"
    fields_desc = [IntField("seconds", 0), IntField("nanoseconds", 0)]

    def extract_padding(self, s):
        return b"", s


class APS(Packet):
    """
    Linear protective switching APS data packet
    """

    name = "APS"

    fields_desc = [
        BitEnumField(
            "req_st",
            0,
            4,
            {
                0b0000: "No request (NR)",
                0b0001: "Do not request (DNR)",
                0b0010: "Reverse request (RR)",
                0b0100: "Exercise (EXER)",
                0b0101: "Wait-to-restore (WTR)",
                0b0110: "Deprecated",
                0b0111: "Manual switch (MS)",
                0b1001: "Signal degrade (SD)",
                0b1011: "Signal fail for working (SF)",
                0b1101: "Forced switch (FS)",
                0b1110: "Signal fail on protection (SF-P)",
                0b1111: "Lockout of protection (LO)",
            },
        ),
        FlagsField(
            "prot_type",
            0,
            4,
            {
                (1 << 3): "A",
                (1 << 2): "B",
                (1 << 1): "D",
                (1 << 0): "R",
            },
        ),
        EnumField(
            "req_sig", 0, {0: "Null signal", 1: "Normal traffic"}, fmt="B"
        ),
        EnumField(
            "br_sig", 0, {0: "Null signal", 1: "Normal traffic"}, fmt="B"
        ),
        FlagsField("br_type", 0, 8, {(1 << 7): "T"}),
    ]

    def extract_padding(self, s):
        return b"", s


class RAPS(Packet):
    """
    Ring protective switching R-APS data packet
    """

    name = "R-APS"

    fields_desc = [
        BitEnumField(
            "req_st",
            0,
            4,
            {
                0b0000: "No request (NR)",
                0b0111: "Manual switch (MS)",
                0b1011: "Signal fail(SF)",
                0b1101: "Forced switch (FS)",
                0b1110: "Event",
            },
        ),
        MultipleTypeField(
            [
                (
                    BitEnumField("sub_code", 0, 4, {0b0000: "Flush"}),
                    lambda p: p.req_st == 0b1110,
                )
            ],
            BitField("sub_code", 0, 4),
        ),
        FlagsField(
            "status",
            0,
            8,
            {
                (1 << 7): "RB",
                (1 << 6): "DNF",
                (1 << 5): "BPR",
            },
        ),
        MACField("node_id", None),
        NBytesField("resv", 0, 24),
    ]

    def extract_padding(self, s):
        return b"", s


class OAM(Packet):
    """
    OAM data unit
    """

    name = "OAM"

    OPCODES = {
        1: "Continuity Check Message (CCM)",
        3: "Loopback Message (LBM)",
        2: "Loopback Reply (LBR)",
        5: "Linktrace Message (LTM)",
        4: "Linktrace Reply (LTR)",
        32: "Generic Notification Message (GNM)",
        33: "Alarm Indication Signal (AIS)",
        35: "Lock Signal (LCK)",
        37: "Test Signal (TST)",
        39: "Automatic Protection Switching (APS)",
        40: "Ring-Automatic Protection Switching (R-APS)",
        41: "Maintenance Communication Channel (MCC)",
        43: "Loss Measurement Message (LMM)",
        42: "Loss Measurement Reply (LMR)",
        45: "One Way Delay Measurement (1DM)",
        47: "Delay Measurement Message (DMM)",
        46: "Delay Measurement Reply (DMR)",
        49: "Experimental OAM Message (EXM)",
        48: "Experimental OAM Reply (EXR)",
        51: "Vendor Specific Message (VSM)",
        50: "Vendor Specific Reply (VSR)",
        52: "Client Signal Fail (CSF)",
        53: "One Way Synthetic Loss Measurement (1SL)",
        55: "Synthetic Loss Message (SLM)",
        54: "Synthetic Loss Reply (SLR)",
    }

    TIME_FLAGS = {
        0b000: "Invalid value",
        0b001: "Trans Int 3.33ms",
        0b010: "Trans Int 10ms",
        0b011: "Trans Int 100ms",
        0b100: "Trans Int 1s",
        0b101: "Trans Int 10s",
        0b110: "Trans Int 1min",
        0b111: "Trans Int 10min",
    }

    PERIOD_FLAGS = {
        0b100: "1 frame per second",
        0b110: "1 frame per minute",
    }

    BNM_PERIOD_FLAGS = {
        0b100: "1 frame per second",
        0b101: "1 frame per 10 seconds",
        0b110: "1 frame per minute",
    }

    fields_desc = [
        # Common fields
        BitField("mel", 0, 3),
        MultipleTypeField(
            [(BitField("version", 1, 5), lambda x: x.opcode in [43, 45, 47])],
            BitField("version", 0, 5),
        ),
        EnumField("opcode", None, OPCODES, fmt="B"),
        MultipleTypeField(
            [
                (
                    FlagsField("flags", 0, 5, {(1 << 4): "RDI"}),
                    lambda x: x.opcode == 1,
                ),
                (
                    FlagsField("flags", 0, 8, {(1 << 7): "HWonly"}),
                    lambda x: x.opcode == 5,
                ),
                (
                    FlagsField(
                        "flags",
                        0,
                        8,
                        {
                            (1 << 7): "HWonly",
                            (1 << 6): "FwdYes",
                            (1 << 5): "TerminalMEP",
                        },
                    ),
                    lambda x: x.opcode == 4,
                ),
                (BitField("flags", 0, 5), lambda x: x.opcode in [33, 35, 32]),
                (
                    FlagsField("flags", 0, 8, {1: "Proactive"}),
                    lambda x: x.opcode in [43, 45, 47],
                ),
                (
                    BitEnumField(
                        "flags",
                        0,
                        5,
                        {
                            0b000: "LOS",
                            0b001: "FDI",
                            0b010: "RDI",
                            0b011: "DCI",
                        },
                    ),
                    lambda x: x.opcode == 52,
                ),
            ],
            ByteField("flags", 0),
        ),
        ConditionalField(
            MultipleTypeField(
                [
                    (
                        BitEnumField("period", 1, 3, TIME_FLAGS),
                        lambda x: x.opcode == 1,
                    ),
                    (
                        BitEnumField("period", 0b110, 3, BNM_PERIOD_FLAGS),
                        lambda x: x.opcode in [13, 32],
                    ),
                ],
                BitEnumField("period", 0b110, 3, PERIOD_FLAGS),
            ),
            lambda x: x.opcode in [1, 33, 35, 52, 32],
        ),
        MultipleTypeField(
            [
                (ByteField("tlv_offset", 70), lambda x: x.opcode == 1),
                (
                    ByteField("tlv_offset", 4),
                    lambda x: x.opcode in [3, 2, 37, 39],
                ),
                (ByteField("tlv_offset", 17), lambda x: x.opcode == 5),
                (ByteField("tlv_offset", 6), lambda x: x.opcode == 4),
                (ByteField("tlv_offset", 32), lambda x: x.opcode in [40, 47]),
                (ByteField("tlv_offset", 12), lambda x: x.opcode == 43),
                (
                    ByteField("tlv_offset", 16),
                    lambda x: x.opcode in [45, 54, 53, 55],
                ),
                (ByteField("tlv_offset", 13), lambda x: x.opcode == 32),
                (
                    ByteField("tlv_offset", 10),
                    lambda x: x.opcode == 41
                ),
            ],
            ByteField("tlv_offset", 0),
        ),
        # End common fields
        ConditionalField(
            IntField("seq_num", 0), lambda x: x.opcode in [1, 3, 2, 37]
        ),
        ConditionalField(IntField("trans_id", 0),
                         lambda x: x.opcode in [5, 4]),
        ConditionalField(
            OUIField("oui", None), lambda x: x.opcode in [41, 49, 48, 51, 50]
        ),
        ConditionalField(
            MultipleTypeField(
                [(ByteField("subopcode", 1), lambda x: x.opcode == 32)],
                ByteField("subopcode", 0),
            ),
            lambda x: x.opcode in [41, 49, 48, 51, 50, 32],
        ),
        ConditionalField(
            MepIdField("mep_id", 0),
            lambda x: x.opcode == 1 \
            or (x.opcode == 41 and x.subopcode == 1 and x.oui == 6567),
        ),
        ConditionalField(
            PacketField("meg_id", MegId(), MegId), lambda x: x.opcode == 0x01
        ),
        ConditionalField(
            ShortField("src_mep_id", 0), lambda x: x.opcode in [55, 54, 53]
        ),
        ConditionalField(
            ShortField("rcv_mep_id", 0), lambda x: x.opcode in [55, 54, 53]
        ),
        ConditionalField(
            IntField("test_id", 0), lambda x: x.opcode in [55, 54, 53]
        ),
        ConditionalField(
            IntField("txfcf", 0), lambda x: x.opcode in [1, 43, 42, 55, 54, 53]
        ),
        ConditionalField(IntField("rxfcb", 0), lambda x: x.opcode == 1),
        ConditionalField(IntField("rxfcf", 0), lambda x: x.opcode in [43, 42]),
        ConditionalField(
            IntField("txfcb", 0), lambda x: x.opcode in [1, 43, 42, 55, 54]
        ),
        ConditionalField(IntField("resv", 0), lambda x: x.opcode in [1, 53]),
        ConditionalField(ByteField("ttl", 0), lambda x: x.opcode in [5, 4]),
        ConditionalField(MACField("orig_mac", None), lambda x: x.opcode == 5),
        ConditionalField(MACField("targ_mac", None), lambda x: x.opcode == 5),
        ConditionalField(ByteField("relay_act", None),
                         lambda x: x.opcode == 4),
        ConditionalField(
            PacketField("txtsf", PTP_TIMESTAMP(), PTP_TIMESTAMP),
            lambda x: x.opcode in [45, 47, 46],
        ),
        ConditionalField(
            PacketField("rxtsf", PTP_TIMESTAMP(), PTP_TIMESTAMP),
            lambda x: x.opcode in [45, 47, 46],
        ),
        ConditionalField(
            PacketField("txtsb", PTP_TIMESTAMP(), PTP_TIMESTAMP),
            lambda x: x.opcode in [47, 46],
        ),
        ConditionalField(
            PacketField("rxtsb", PTP_TIMESTAMP(), PTP_TIMESTAMP),
            lambda x: x.opcode in [47, 46],
        ),
        ConditionalField(
            IntField("expct_dur", None),
            lambda x: x.opcode == 41 and x.subopcode == 1 and x.oui == 6567,
        ),
        ConditionalField(IntField("nom_bdw", None), lambda x: x.opcode == 32),
        ConditionalField(IntField("curr_bdw", None), lambda x: x.opcode == 32),
        ConditionalField(IntField("port_id", None), lambda x: x.opcode == 32),
        ConditionalField(
            PacketField("aps", APS(), APS), lambda x: x.opcode == 39
        ),
        ConditionalField(
            PacketField("raps", RAPS(), RAPS), lambda x: x.opcode == 40
        ),
        ConditionalField(
            PacketListField("tlvs", [], next_cls_cb=guess_tlv_type),
            lambda x: x.opcode in [3, 2, 5, 4, 37, 45, 47, 46, 55, 54, 53],
        ),
        ConditionalField(
            IntField("opt_data", None),
            lambda x: x.opcode in [49, 48, 51, 50] and False,
        ),  # FIXME: field documented elsewhere
        # TODO: add EXM, EXR, VSM and VSR data
        ByteField("end_tlv", 0),
    ]


bind_layers(Dot1Q, OAM, type=0x8902)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Carmine Scarpitta <carmine.scarpitta@uniroma2.it>

# scapy.contrib.description = Simple Two-Way Active Measurement Protocol (STAMP)
# scapy.contrib.status = loads

"""
STAMP (Simple Two-Way Active Measurement Protocol) - RFC 8762.

References:
    * `Simple Two-Way Active Measurement Protocol [RFC 8762]
      <https://www.rfc-editor.org/rfc/rfc8762.html>`_
    * `Simple Two-Way Active Measurement Protocol Optional Extensions [RFC 8972]
      <https://www.rfc-editor.org/rfc/rfc8972.html>`_
"""

from scapy import config
from scapy.base_classes import Packet_metaclass
from scapy.layers.inet import UDP
from scapy.layers.ntp import TimeStampField
from scapy.packet import Packet, bind_layers
from scapy.fields import (
    BitEnumField,
    BitField,
    ByteEnumField,
    ByteField,
    FlagsField,
    IntField,
    MultipleTypeField,
    NBytesField,
    PacketField,
    PacketListField,
    ShortField,
    StrLenField,
    UTCTimeField
)


_sync_types = {
    0: 'No External Synchronization for the Time Source',
    1: 'Clock Synchronized to UTC using an External Source'
}

_timestamp_types = {
    0: 'NTP 64-bit Timestamp Format',
    1: 'PTPv2 Truncated Timestamp Format'
}

_stamp_tlvs = {

}


class ErrorEstimate(Packet):
    """
    The Error Estimate specifies the estimate of the error and
    synchronization. The format of the Error Estimate field
    (defined in Section 4.1.2 of `RFC 4656
    <https://www.rfc-editor.org/rfc/rfc4656.html>`_) is reported below::

        0                   1
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |S|Z|   Scale   |   Multiplier  |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    ``S`` is interpreted as follows:

        +-------+-------------------------------------------------------+
        | Value | Description                                           |
        +-------+-------------------------------------------------------+
        |   0   | there is no notion of external synchronization for    |
        |       | the time source                                       |
        +-------+-------------------------------------------------------+
        |   1   | the party generating the timestamp has a clock that   |
        |       | is synchronized to UTC using an external source       |
        +-------+-------------------------------------------------------+

    ``Z`` is interpreted as follows (defined in Section 2.3 of `RFC 8186
    <https://www.rfc-editor.org/rfc/rfc8186.html>`_):

        +-------+---------------------------------------+
        | Value | Description                           |
        +-------+---------------------------------------+
        |   0   | NTP 64-bit format of a timestamp      |
        +-------+---------------------------------------+
        |   1   | PTPv2 truncated format of a timestamp |
        +-------+---------------------------------------+

    ``Scale`` and ``Multiplier`` are linked by the following relationship::

        ErrorEstimate = Multiplier*2^(-32)*2^Scale (in seconds)


    References:
        * `A One-way Active Measurement Protocol (OWAMP) [RFC 4656]
          <https://www.rfc-editor.org/rfc/rfc4656.html>`_
        * `Support of the IEEE 1588 Timestamp Format in a Two-Way Active
          Measurement Protocol (TWAMP) [RFC 8186]
          <https://www.rfc-editor.org/rfc/rfc8186.html>`_
    """

    name = 'Error Estimate'
    fields_desc = [
        BitEnumField('S', 0, 1, _sync_types),
        BitEnumField('Z', 0, 1, _timestamp_types),
        BitField('scale', 0, 6),
        ByteField('multiplier', 1),
    ]

    def guess_payload_class(self, payload):
        # type: (str) -> Packet_metaclass
        # Trick to tell scapy that the remaining bytes of the currently
        # dissected string is not a payload of this packet but of some other
        # underlayer packet
        return config.conf.padding_layer


class STAMPTestTLV(Packet):
    """
    The STAMP Test TLV defined in Section 4 of [RFC 8972] provides a flexible
    extension mechanism for optional informational elements.

    The TLV Format in a STAMP Test packet is reported below::

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |STAMP TLV Flags|     Type      |           Length              |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        ~                            Value                              ~
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


    +-------+---------+-------------------------------------------------+
    | Field           | Description                                     |
    +-----------------+-------------------------------------------------+
    | STAMP TLV Flags | 8-bit field; for the details about the STAMP    |
    |                 | TLV Flags Format, see RFC 8972                  |
    +-----------------+-------------------------------------------------+
    | Type            | characterizes the interpretation of the Value   |
    |                 | field                                           |
    +-----------------+-------------------------------------------------+
    | Length          | the length of the Value field in octets         |
    +-----------------+-------------------------------------------------+
    | Value           | interpreted according to the value of the Type  |
    |                 | field                                           |
    +-----------------+-------------------------------------------------+


    References:
        * `Simple Two-Way Active Measurement Protocol Optional Extensions
          [RFC 8972] <https://www.rfc-editor.org/rfc/rfc8972.html>`_
    """

    name = 'STAMP Test Packet - Generic TLV'
    fields_desc = [
        FlagsField('flags', 0, 8, "UMIRRRRR"),
        ByteEnumField('type', None, _stamp_tlvs),
        ShortField('len', 0),
        StrLenField('value', '', length_from=lambda pkt: pkt.len),
    ]

    def extract_padding(self, p):
        return b"", p

    registered_stamp_tlv = {}

    @classmethod
    def register_variant(cls):
        cls.registered_stamp_tlv[cls.type.default] = cls

    @classmethod
    def dispatch_hook(cls, pkt=None, *args, **kargs):
        if pkt:
            tmp_type = ord(pkt[1:2])
            return cls.registered_stamp_tlv.get(tmp_type, cls)
        return cls


class STAMPSessionSenderTestUnauthenticated(Packet):
    """
    Extended STAMP Session-Sender Test Packet in Unauthenticated Mode.

    The format (defined in Section 3 of `RFC 8972
    <https://www.rfc-editor.org/rfc/rfc8972.html>`_) is shown below::

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                        Sequence Number                        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          Timestamp                            |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |         Error Estimate        |             SSID              |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        |                                                               |
        |                         MBZ (28 octets)                       |
        |                                                               |
        |                                                               |
        |                                                               |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        ~                            TLVs                               ~
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    References:
        * `Simple Two-Way Active Measurement Protocol Optional Extensions
          [RFC 8972] <https://www.rfc-editor.org/rfc/rfc8972.html>`_
    """
    name = 'STAMP Session-Sender Test'
    fields_desc = [
        IntField('seq', 0),
        MultipleTypeField(
            [
                (TimeStampField('ts', 0),
                    lambda pkt:pkt.err_estimate.Z == 0)
            ],
            UTCTimeField('ts', 0, fmt='Q')
        ),
        PacketField('err_estimate', ErrorEstimate(), ErrorEstimate),
        ShortField('ssid', 1),
        NBytesField('mbz', 0, 28),  # 28 bytes MBZ
        PacketListField('tlv_objects', [], STAMPTestTLV),
    ]


class STAMPSessionReflectorTestUnauthenticated(Packet):
    """
    Extended STAMP Session-Reflector Test Packet in Unauthenticated Mode.

    The format (defined in Section 3 of `RFC 8972
    <https://www.rfc-editor.org/rfc/rfc8972.html>`_) is shown below::

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                        Sequence Number                        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          Timestamp                            |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |         Error Estimate        |           SSID                |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          Receive Timestamp                    |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                 Session-Sender Sequence Number                |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                  Session-Sender Timestamp                     |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Session-Sender Error Estimate |           MBZ                 |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |Ses-Sender TTL |                   MBZ                         |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        ~                            TLVs                               ~
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    References:
        * `Simple Two-Way Active Measurement Protocol Optional Extensions
          [RFC 8972] <https://www.rfc-editor.org/rfc/rfc8972.html>`_
    """
    name = 'STAMP Session-Reflector Test'
    fields_desc = [
        IntField('seq', 0),
        MultipleTypeField(
            [
                (TimeStampField('ts', 0),
                    lambda pkt:pkt.err_estimate.Z == 0),
            ],
            UTCTimeField('ts', 0, fmt='Q')
        ),
        PacketField('err_estimate', ErrorEstimate(), ErrorEstimate),
        ShortField('ssid', 1),
        MultipleTypeField(
            [
                (TimeStampField('ts_rx', 0),
                    lambda pkt:pkt.err_estimate.Z == 0)
            ],
            UTCTimeField('ts_rx', 0, fmt='Q')
        ),
        IntField('seq_sender', 0),
        MultipleTypeField(
            [
                (TimeStampField('ts_sender', 0),
                    lambda pkt:pkt.err_estimate_sender.Z == 0)
            ],
            UTCTimeField('ts_sender', 0, fmt='Q')
        ),
        PacketField('err_estimate_sender', ErrorEstimate(), ErrorEstimate),
        ShortField('mbz1', 0),
        ByteField('ttl_sender', 255),
        NBytesField('mbz2', 0, 3),  # 3 bytes MBZ
        PacketListField('tlv_objects', [], STAMPTestTLV),
    ]


bind_layers(UDP, STAMPSessionSenderTestUnauthenticated, dport=862)
bind_layers(UDP, STAMPSessionReflectorTestUnauthenticated, sport=862)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Ethernet Synchronization Message Channel (ESMC)
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers
from scapy.fields import BitField, ByteField, XByteField, ShortField, XStrFixedLenField  # noqa: E501
from scapy.contrib.slowprot import SlowProtocol
from scapy.compat import orb


class ESMC(Packet):
    name = "ESMC"
    fields_desc = [
        XStrFixedLenField("ituOui", b"\x00\x19\xa7", 3),
        ShortField("ituSubtype", 1),
        BitField("version", 1, 4),
        BitField("event", 0, 1),
        BitField("reserved1", 0, 3),
        XStrFixedLenField("reserved2", b"\x00" * 3, 3),
    ]

    def guess_payload_class(self, payload):
        if orb(payload[0]) == 1:
            return QLTLV
        if orb(payload[0]) == 2:
            return EQLTLV
        return Packet.guess_payload_class(self, payload)


class QLTLV(ESMC):
    name = "QLTLV"
    fields_desc = [
        ByteField("type", 1),
        ShortField("length", 4),
        XByteField("ssmCode", 0xf),
    ]


class EQLTLV(ESMC):
    name = "EQLTLV"
    fields_desc = [
        ByteField("type", 2),
        ShortField("length", 0x14),
        XByteField("enhancedSsmCode", 0xFF),
        XStrFixedLenField("clockIdentity", b"\x00" * 8, 8),
        ByteField("flag", 0),
        ByteField("cascaded_eEEcs", 1),
        ByteField("cascaded_EEcs", 0),
        XStrFixedLenField("reserved", b"\x00" * 5, 5),
    ]


bind_layers(SlowProtocol, ESMC, subtype=10)
